#include "defaults.h"
#include "cl_config.h"
#include "verbosity.h"
#include "wchar_utils.h"
#include "bunch.h"

#include <errno.h>
#include <wchar.h>
#include <popt.h>
#include <locale.h>
#include <string.h>
#include <limits.h>
#include <wctype.h>
#include <stdlib.h>

#include "mouchard.h"

int main(int argc, const char **argv) {
  /* Parameters */
  static int line_max_length = DEFAULT_MAX_LINE_LENGTH;
  static char *encoding = NULL;
  static int substrings = 0;
  static int words = 0;
  /*static int opt_density = DEFAULT_OPT_DENSITY;*/
  static int dichotomic = 0;


  /* Command line parsing */
  static struct poptOption cl_verbosity_options[] = CL_VERBOSITY_OPTIONS;
  struct poptOption cl_options[] = {
    POPT_AUTOHELP
    { NULL , '\0' , POPT_ARG_INCLUDE_TABLE , cl_verbosity_options , 0 , "Verbosity options:" , NULL },
    { CLO_LONG_MAX_LENGTH     , CLO_SHORT_MAXLENGTH      , POPT_ARG_INT    , &line_max_length    , 0                         , CLO_HELP_TS_MAXLENGTH      , "LENGTH" },
    { CLO_LONG_ENCODING       , CLO_SHORT_ENCODING       , POPT_ARG_STRING , &encoding           , 0                         , CLO_HELP_TS_ENCODING       , "ENCODING" },
    { CLO_LONG_SUBSTRINGS     , CLO_SHORT_SUBSTRINGS     , POPT_ARG_NONE   , &substrings         , 0                         , CLO_HELP_TS_SUBSTRINGS     , NULL },
    { CLO_LONG_WORDS          , CLO_SHORT_WORDS          , POPT_ARG_NONE   , &words              , 0                         , CLO_HELP_TS_WORDS          , NULL },
    { "dichotomic"            , '\0'                     , POPT_ARG_NONE   , &dichotomic         , 0                         , "Perform dichotomic search", NULL },
    POPT_TABLEEND
  };
  poptContext cl_context = poptGetContext(NULL, argc, argv, cl_options, 0);
  int cl_oh;

  /* Local vars */
  const char *out_fn = NULL;
  const char *in_fn = NULL;
  const char *dict_fn = NULL;
  FILE *in_fh = NULL;
  FILE *out_fh = NULL;
  FILE *dict_fh = NULL;
  unsigned long line_nb;
  unsigned long dict_size;
  int nonls = 0;
  wchar_t *line = NULL;
  wchar_t *saveline = NULL;
  wchar_t **dict = calloc(1800000, sizeof(wchar_t *));
  size_t *dict_len = calloc(1800000, sizeof(size_t));
  bunch_t *dict_entries = bunch_new(sizeof(wchar_t), 20000000);


  poptSetOtherOptionHelp(cl_context, "DICTFILE [INFILE [OUTFILE]]");
  while ((cl_oh = poptGetNextOpt(cl_context)) > 0) {
    switch (cl_oh) {
    case CLO_SWITCH_QUIET:
      verbosity = V_QUIET;
      break;
    case CLO_SWITCH_VERBOSE:
      verbosity = V_VERBOSE;
      break;
    case CLO_SWITCH_LOCACE:
      verbosity = V_LOCACE;
      break;
    case CLO_SWITCH_STRICT:
      severity = V_WARNING;
      break;
    }
  }
  if (cl_oh != -1)
    verbosity_message(V_QUIET | V_ERROR, "%s %s", poptStrerror(cl_oh), poptBadOption(cl_context, 0));

  if (poptPeekArg(cl_context)) {
    dict_fn = poptGetArg(cl_context);
  }
  else {
    verbosity_message(V_QUIET | V_ERROR, "You must provide a dictionary file name");
  }

  if (poptPeekArg(cl_context)) {
    in_fn =  poptGetArg(cl_context);
  }

  if (poptPeekArg(cl_context)) {
    out_fn =  poptGetArg(cl_context);
  }

  poptFreeContext(cl_context);



  /* Change locale */
  if (encoding == NULL) {
    encoding = setlocale(LC_CTYPE, NULL);
  }
  else if (setlocale(LC_CTYPE, encoding) == NULL) {
    verbosity_message(V_QUIET | V_ERROR, "Unhandled locale '%s'", encoding);
  }
  verbosity_message(V_NORMAL | V_INFO, "Locale set to '%s'", encoding);

  /* Open files and read dictionary */
  if (in_fn != NULL) {
    verbosity_message(V_VERBOSE | V_INFO, "Opening input file '%s'", in_fn);
    if (!(in_fh = fopen(in_fn, "r"))) {
      verbosity_message(V_QUIET | V_ERROR, "Could not open file '%s'", in_fn);
    }
  }
  else {
    in_fh = stdin;
  }

  if (out_fn != NULL) {
    verbosity_message(V_VERBOSE | V_INFO, "Opening file '%s' for writing", out_fn);
    if (!(out_fh = fopen(out_fn, "w"))) {
      verbosity_message(V_QUIET | V_ERROR, "Could not open '%s' for writing", out_fn);
    }
  }
  else {
    out_fh = stdout;
  }

  verbosity_message(V_NORMAL | V_INFO, "Reading dictionary file '%s'", dict_fn);
  if (!(dict_fh = fopen(dict_fn, "r"))) {
    verbosity_message(V_QUIET | V_ERROR, "Could not open dictionary file '%s'", dict_fn);
  }
  line = calloc(line_max_length + 1, sizeof(wchar_t));
  for (line_nb = 1; fgetws(line, line_max_length, dict_fh); line_nb++) {
    size_t l = wcslen(line);
    if (line_nb == 1)
      nonls = get_nonls(line, l);
    if (l - nonls) {
      line[l - nonls] = L'\0';
      dict_len[line_nb - 1] = l - nonls;
      dict[line_nb - 1] = bunch_claim(dict_entries, l);
      wcscat(dict[line_nb - 1], line);
    }
  }
  if (errno == EILSEQ) {
    verbosity_message(V_NORMAL | V_WARNING, "Illegal character sequence, line %ld", line_nb);
  }
  dict_size = line_nb;


  /* Read input and search */
  verbosity_message(V_NORMAL | V_INFO, "Reading input");
  /*DDD search = psearch_new((td->multi ? td->multimap->pat : td->map), search_options, gci, opt_density);*/
  for (line_nb = 1; fgetws(line, line_max_length, in_fh); line_nb++) {
    size_t l = wcslen(line);
    if (line_nb == 1)
      nonls = get_nonls(line, l);
    if (l - nonls) {
      int alnum = 0;
      line[l - nonls] = L'\0';
      saveline = line;
      do {
	if (!(words && alnum)) {
	  if (dichotomic) {
	    unsigned int lo = 0;
	    unsigned int hi = dict_size - 2;
	    do {
	      unsigned int i = (hi + lo) / 2;
	      int d;
	      d = wcsncmp(saveline, dict[i], dict_len[i]);
	      if ((d == 0) && !(words && (iswalpha(saveline[dict_len[i]])))) {
		print_wchar(out_fh, dict[i], 1);
		break;
	      }
	      else if (lo >= hi) {
		break;
	      }
	      else if (d < 0) {
		hi = i ? i - 1 : 0;
	      }
	      else {
		lo = i + 1;
	      }
	    } while (1);
	  }
	  else {
	    unsigned int i;
	    for (i = 0; i < dict_size - 1; i++)
	      if ((!wcsncmp(saveline, dict[i], dict_len[i])) && !(words && (iswalpha(saveline[dict_len[i]])))) {
		print_wchar(out_fh, dict[i], 1);
		break;
	      }
	  }
	}
	alnum = iswalpha(*saveline);
	saveline++;
      } while (substrings && *saveline);
    }
  }
  if (errno == EILSEQ) {
    verbosity_message(V_NORMAL | V_WARNING, "Illegal character sequence, line %ld", line_nb);
  }

  free(line);

  return 0;
}
