/*
  Copyright (C) 1988-1994,1996-1999,2003,2004 Free Software Foundation, Inc.
  This file is part of the GNU C Library.  Its master source is NOT part of
  the C library, however.  The master source lives in /gd/gnu/lib.
  
  The GNU C Library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.
  
  The GNU C Library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.
  
  You should have received a copy of the GNU Lesser General Public
  License along with the GNU C Library; if not, write to the Free
  Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
  02111-1307 USA.
*/

#include "bunch.h"

#include "mouchard.h"

bunch_t *bunch_new(size_t element_size, unsigned int block_size) {
  bunch_t *b = calloc(1, sizeof(bunch_t));

  b->element_size = element_size;
  b->block_size = block_size;

  /* Start with allocating a single block */
  b->block_number = 1;
  b->blocks = calloc(1, sizeof(bunch_block_t *));
  b->blocks[0] = calloc(1, sizeof(bunch_block_t));
  b->blocks[0]->unclaimed = block_size;
  b->blocks[0]->data = calloc(block_size, element_size);
  b->attrib = BUNCH_ATTRIB_BLOCK_ORDER_UNCLAIMED;

  return b;
}

void bunch_free(bunch_t *b) {
  int i;

  for (i = 0; i < b->block_number; i++) {
    free(b->blocks[i]->data);
    free(b->blocks[i]);
  }
  free(b->blocks);
  free(b);
}


/*
  search_best_block - searches for the best block to claim objects
  b - bunch where objects have been claimed
  x - size of the array claimed
  return value - index of the best block

  Note: this function looks for the block with the less unclaimed objects but more than x, it assumes that ther is at least one block with at least x unclaimed objects, it also assumes that blocks are in ascending order of unclaimed objects (dichotomic search)
 */

int search_best_block(bunch_t *b, size_t x) {
  int lo = 0;
  int hi = b->block_number - 1;

  do {
    if (lo == hi)
      return lo;
    else {
      int av = (lo + hi) / 2;
      if (x > b->blocks[av]->unclaimed)
	lo = av + 1;
      else if (x < b->blocks[av]->unclaimed)
	hi = av;
      else
	return av;
    }
  } while (1);
}


/*
  block_insert_sort - sorts a block in ascending order of unclaimed objects
  b - bunch to sort
  i - index of the block to put in place (will look from i downto 0)

  Note: this is insert sort, so before you get started, if you have performance problems, the YOU should set a larger block_size in order to have less blocks
 */

void block_insert_sort(bunch_t *b, int i) {
  bunch_block_t *p = b->blocks[i];
  unsigned int vi = p->unclaimed;
  int j;
  for (j = i - 1; j >= 0; j--) {
    unsigned int vj = b->blocks[j]->unclaimed;
    if (vj > vi) {
      b->blocks[j + 1] = b->blocks[j];
    }
    else {
      break;
    }
  }
  b->blocks[j + 1] = p;
}


void *bunch_claim(bunch_t *b, size_t nb) {
  void *p = NULL;
  int best_block;
  bunch_block_t *best_block_ptr;

  /* Avoid strange cases */
  if (!nb) {
    return NULL;
  }

  if ((b->attrib & BUNCH_ATTRIB_PRUNED_BLOCKS) || (nb > b->block_size)) {
    return NULL;
  }

  /* Sort blocks in ascending order of unclaimed objects if necessary */
  if (!(b->attrib & BUNCH_ATTRIB_BLOCK_ORDER_UNCLAIMED)) {
    int i;
    for (i = 1; i < b->block_number; i++)
      block_insert_sort(b, i);
    b->attrib |= BUNCH_ATTRIB_BLOCK_ORDER_UNCLAIMED;
  }

  if (b->blocks[b->block_number - 1]->unclaimed < nb) {
    /* Last block has not enough unclaimed objects, allocate a new block */
    best_block = b->block_number;
    b->block_number++;
    b->blocks = realloc(b->blocks, sizeof(bunch_block_t *) * b->block_number);
    best_block_ptr = calloc(1, sizeof(bunch_block_t));
    best_block_ptr->unclaimed = b->block_size;
    best_block_ptr->data = calloc(b->block_size, b->element_size);
    b->blocks[best_block] = best_block_ptr;
  }
  else {
    /* Otherwise choose the best block */
    best_block = search_best_block(b, nb);
    best_block_ptr = b->blocks[best_block];
  }

  /* Get the pointer to the new array */
  p = (void *)((intptr_t)best_block_ptr->data + (b->block_size - best_block_ptr->unclaimed) * b->element_size);

  /* Update unclaimed number of objects and reorder the block */
  best_block_ptr->unclaimed -= nb;
  block_insert_sort(b, best_block);

  return p;
}


void bunch_save(bunch_t *b, int options, FILE *f) {
  int i, attrib = b->attrib;

  if (options & BUNCH_ATTRIB_PRUNED_BLOCKS)
    b->attrib = attrib | BUNCH_ATTRIB_PRUNED_BLOCKS;

  /* Write bunch */
  fwrite(b, sizeof(bunch_t), 1, f);

  /* Write each bunch */
  for (i = 0; i < b->block_number; i++) {
    fwrite(b->blocks[i], sizeof(bunch_block_t), 1, f);
    if (options & BUNCH_ATTRIB_PRUNED_BLOCKS)
      /* Prune block, ie only write claimed blocks */
      fwrite(b->blocks[i]->data, b->element_size, b->block_size - b->blocks[i]->unclaimed, f);
    else
      fwrite(b->blocks[i]->data, b->element_size, b->block_size, f);
  }

  /* Restore original attributes */
  b->attrib = attrib;
}


bunch_t *bunch_load(FILE *f) {
  bunch_t *b = calloc(1, sizeof(bunch_t));
  void *p;
  int i;

  /* Read bunch */
  fread(b, sizeof(bunch_t), 1, f);

  /* Read blocks */
  b->blocks = calloc(b->block_number, sizeof(bunch_block_t *));
  for (i = 0; i < b->block_number; i++) {
    b->blocks[i] = calloc(1, sizeof(bunch_block_t));
    fread(b->blocks[i], sizeof(bunch_block_t), 1, f);
    if (b->attrib & BUNCH_ATTRIB_PRUNED_BLOCKS) {
      /* If pruned, read only claimed objects */
      p = calloc(b->block_size - b->blocks[i]->unclaimed, b->element_size);
      fread(p, b->element_size, b->block_size - b->blocks[i]->unclaimed, f);
    }
    else {
      p = calloc(b->block_size, b->element_size);
      fread(p, b->element_size, b->block_size, f);
    }

    /* Update offset for bunch_correct_ptr */
    b->blocks[i]->offset = (intptr_t)p - (intptr_t)b->blocks[i]->data;
    b->blocks[i]->data = p;
  }

  return b;
}


/*
  search_ptr_block - looks for the block to which belonged an address
  b - bunch where to search
  p - address to search
  return value - index of the block

  Note: this will search for the block such as p was between its start and end addresses, it assumes ther is such a block and that blocks are in ascending order of start addresses (dichotomic search), otherwise the result is undefined
 */

int search_ptr_block(bunch_t *b, void *p) {
  int lo = 0;
  int hi = b->block_number - 1;
  
  do {
    if (hi == lo)
      return lo;
    else {
      int av = 1 + (lo + hi) / 2;
      if ((intptr_t)p > (intptr_t)b->blocks[av]->data - b->blocks[av]->offset) {
	lo = av;
      }
      else if ((intptr_t)p < (intptr_t)b->blocks[av]->data - b->blocks[av]->offset)
	hi = av - 1;
      else
	return av;
    }
  } while (1);
}

void *bunch_correct_ptr(bunch_t *b, void *p) {
  int block;

  /* Avoid weird cases XXX what if the pointer was not from this bunch ? */
  if (p == NULL)
    return NULL;

  /* Sort blocks in ascending order of data address, if necessary */
  /* This is insert sort because I'm a bastard */
  if (b->attrib & BUNCH_ATTRIB_BLOCK_ORDER_UNCLAIMED) {
    int i;
    for (i = 1; i < b->block_number; i++) {
      bunch_block_t *p = b->blocks[i];
      intptr_t vi = (intptr_t)p->data - p->offset;
      int j;
      for (j = i - 1; j >= 0; j--) {
	intptr_t vj = (intptr_t)b->blocks[j]->data - b->blocks[j]->offset;
	if (vj > vi) {
	  b->blocks[j + 1] = b->blocks[j];
	}
	else {
	  break;
	}
      }
      b->blocks[j + 1] = p;
    }
    b->attrib &= ~BUNCH_ATTRIB_BLOCK_ORDER_UNCLAIMED;
    }

  block = search_ptr_block(b, p);
  p = (void *)((intptr_t)p + b->blocks[block]->offset);

  return p;
}


unsigned int bunch_unclaimed(bunch_t *b) {
  unsigned int i, t = 0;

  if (b->attrib & BUNCH_ATTRIB_PRUNED_BLOCKS)
    return 0;

  for (i = 0; i < b->block_number; i++)
    t += b->blocks[i]->unclaimed;

  return t;
}
