/*
  bunch.h - allocation of several contiguous blocks of memory, part of libtrish

  Copyright (C) 1988-1994,1996-1999,2003,2004 Free Software Foundation, Inc.
  This file is part of the GNU C Library.  Its master source is NOT part of
  the C library, however.  The master source lives in /gd/gnu/lib.
  
  The GNU C Library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.
  
  The GNU C Library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.
  
  You should have received a copy of the GNU Lesser General Public
  License along with the GNU C Library; if not, write to the Free
  Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
  02111-1307 USA.
*/

/*
  Summary:

  The bunch_t structure allows to allocate several pointers of the same size at once. The provided bunch_* functions guarantee that the object won't move. The advantages are:
  - one allocation instead of a lot makes faster memory management;
  - contiguous objects makes them easy to dump into a binary file.


  How does it work?

  A bunch allocates objects by blocks of a given size, then you "claim" pointers from the allocated blocks. The bunch keeps an array of blocks with the number of remaining unclaimed objects. When you claim a pointer to N objects, it will look for a block that has enough remaining objects and returns a pointer to an array of N objects. This has three consequences:
  - each claim ensures the N objects are contiguous;
  - you cannot claim more objects than the block size;
  - there will be more objects allocated than claimed, some memory will be waisted.

  Thus it will leave you the delicate exercise to estimate a good block size: large enough to allow larger arrays, but not too large there's not too much waisted memory. There are two mechanisms to contain the consequences of waisted allocation. The bunch_print_info function will give a summary of allocated objects, claimed objects and so on. Also the bunch_save function, used to dump the bunch into a file, may prune unclaimed objects. However if you load a pruned bunch, you cannot claim pointers anymore.


  Step by step:

  Creating:
  1) create a new bunch with bunch_new;
  2) claim pointers as you wish with bunch_claim;
  3) use objects as usual
  4) write the bunch into a file with bunch_save;
  5) free the bunch with bunch_free, this will also free all claimed pointers.

  Loading:
  1) read a bunch from a file;
  2) if the bunch is not pruned, claim more pointers and write again (preferably in another file);
  3) correct pointers (see below);
  4) use objects as usual;
  5) free the bunch with bunch_free, this will also free all claimed pointers.


  Correcting pointers to an object:

  If objects of a bunch contain pointers to other objects from the same bunch (linked lists, trees, etc.), you have to correct those pointers when you load the back in memory. Indeed bunch_load cannot guarantee the objects have the exact same address. Fortunately the bunch keeps the old address and computes the offset from the saved address and the current one. Thus you can (and must) correct pointers to objects from a bunch with bunch_correct_ptr.

*/

#ifndef HAVE_BUNCH_H
#define HAVE_BUNCH_H

#include <stdint.h>
#include <stdlib.h>

/*
  bunch_block_t -  A new bunch_block_t is allocated when you claim more objects than available.
*/
typedef struct {
  /* offset - this is where the offset between the saved address and the current one is kept, only to be used by bunch_correct_ptr */
  intptr_t offset;

  /* unclaimed - number of unclaimed objects */
  unsigned int unclaimed;

  /* data - this is where the objects are held */
  void *data;
} bunch_block_t;


/*
  bunch_t - Bunch main structure, bunch_* functions use a pointer to a bunch_t
*/
typedef struct {
  /* element_size - size of the objects */
  size_t element_size;

  /* block_size - number of objects allocated in each block */
  int block_size;

  /* block_number - number of blocks allocated */
  int block_number;

  /* blocks - array of blocks */
  bunch_block_t **blocks;

  /* attrib - information on the state of the bunch for internal use */
  int attrib;
} bunch_t;


/*
  BUNCH_ATTRIB_* - attrib values
*/

#define BUNCH_ATTRIB_BLOCK_ORDER_UNCLAIMED 1
#define BUNCH_ATTRIB_PRUNED_BLOCKS         2

/*
  bunch_new - creates a new bunch
  element_size - size of the objects usually given by sizeof()
  block_size - number of objects per block
  return value - a freshly allocated bunch
*/
bunch_t *bunch_new(size_t element_size, unsigned int block_size);


/*
  bunch_free - destroys a bunch and frees all claimed pointers
  b - bunch to destroy, it must have been allocated by bunch_new
 */

void bunch_free(bunch_t *b);


/*
  bunch_claim - claims an array of objects
  b - bunch to claim pointers from
  nb - number of consecutive objects desired
  return value - a pointer to an array of objects of size element_size, NULL if nb is greater than block_size or if the bunch is pruned

  Note: DO NOT free a claimed pointer, you may get a segfault, bunch_free will do it cleanly for you
*/

void *bunch_claim(bunch_t *b, size_t nb);

#include <stdio.h>


/*
  bunch_save - writes a bunch into a file
  b - bunch to write
  options - 0 or BUNCH_ATTRIB_PRUNED_BLOCKS to prune the bunch, note that the bunch is not pruned by itself, only the bunch writen on the file (and thus loaded afterwards) will be
  f - file handle, it must be created by fopen with mode "w" (or "a")
*/

void bunch_save(bunch_t *b, int options, FILE *f);


/*
  bunch_load - loads a bunch from a file
  f - file handle, it must be created by fopen with mode "r"
  return value - a new bunch (no need to allocate it)

  Note: it's up to you to place the file handle cursor at the right place when reading a bunch, otherwise the behaviour is undefined
*/
bunch_t *bunch_load(FILE *f);


/*
  bunch_correct_ptr - correct a pointer from a reloaded bunch
  b - bunch from where the pointer was claimed
  p - address to an object from the bunch when it was saved
  return value - the new addres, NULL if p was NULL of not in the range of expected addresses (ie the address was probably not claimed from this bunch)
*/
void *bunch_correct_ptr(bunch_t *b, void *p);
#define BUNCH_CORRECT_PTR(b, p) (p) = bunch_correct_ptr((b), (p))


/*
  bunch_unclaimed - sums the number of unclaimed allocated objects
  b - bunch to compute
  return value - the number of objects allocated but unclaimed

  Note: if the bunch is pruned, this function returns 0
 */

unsigned int bunch_unclaimed(bunch_t *b);

#define BUNCH_CLAIMED(b) ((b)->block_number * (b)->block_size - bunch_unclaimed((b)))

#endif
