/*
  This file is part of the GNU C Library.  Its master source is NOT part of
  the C library, however.  The master source lives in /gd/gnu/lib.

  The GNU C Library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  The GNU C Library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with the GNU C Library; if not, write to the Free
  Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
  02111-1307 USA.
*/

#include "patricia.h"
#include "limits.h"

#include <stdio.h>
#include <wchar.h>
#include <string.h>

#include "mouchard.h"


patricia_t *patricia_new(unsigned int node_block, unsigned int string_block, bunch_t *data) {
  patricia_t *pat = calloc(1, sizeof(patricia_t));
  pat->nodes = bunch_new(sizeof(pnode_t), node_block);
  pat->root = bunch_claim(pat->nodes, 1);
  pat->follow_strings = bunch_new(sizeof(wchar_t), string_block);
  pat->data = data;
  return pat;
}


void patricia_free(patricia_t *pat) {
  bunch_free(pat->nodes);
  bunch_free(pat->follow_strings);
  free(pat);
}

void patricia_save(patricia_t *pat, int options, FILE *f) {
  fwrite(pat, sizeof(patricia_t), 1, f);
  bunch_save(pat->nodes, options, f);
  bunch_save(pat->follow_strings, options, f);
  if (pat->data != NULL)
    bunch_save(pat->data, options, f);
}


patricia_t *patricia_load(FILE *f) {
  int i;
  patricia_t *pat = calloc(1, sizeof(patricia_t));

  fread(pat, sizeof(patricia_t), 1, f);
  pat->nodes = bunch_load(f);
  pat->follow_strings = bunch_load(f);
  if (pat->data != NULL)
    pat->data = bunch_load(f);

  /* Correct pnode_t pointer members */
  for (i = 0; i < pat->nodes->block_number; i++) {
    int j;
    for (j = 0; j < pat->nodes->block_size - pat->nodes->blocks[i]->unclaimed; j++) {
      pnode_t *pn = (pnode_t *)pat->nodes->blocks[i]->data + j;
      BUNCH_CORRECT_PTR(pat->follow_strings, pn->follow_string);
      BUNCH_CORRECT_PTR(pat->nodes, pn->next);
      BUNCH_CORRECT_PTR(pat->nodes, pn->left);
      BUNCH_CORRECT_PTR(pat->nodes, pn->right);
      BUNCH_CORRECT_PTR(pat->nodes, pn->parent);
      if (pat->data != NULL)
	BUNCH_CORRECT_PTR(pat->data, pn->data);
    }
  }

  pat->root = bunch_correct_ptr(pat->nodes, pat->root);

  return pat;
}


void patricia_add_context_init(patricia_add_context_t *ctx, patricia_t *pat, const wchar_t *word, int options, void *arg) {
  ctx->pat = pat;
  ctx->options = options;
  ctx->word = word;
  ctx->word_pos = 0;
  ctx->current = pat->root;
  ctx->follow_pos = 0;
  ctx->event = 0;
  ctx->parent = NULL;
  ctx->arg = arg;
}


#define PATRICIA_ADD_HOOK { if (hook != NULL) hook(ctx); }
void patricia_add(patricia_add_context_t *ctx, void (*hook)(const patricia_add_context_t *)) {

  /* Add start of word node if required */
  if (ctx->options & PATRICIA_ADD_OPTION_START) {
    int is_new = pnode_insert(ctx->pat->nodes, ctx->pat->root, PNODE_TYPE_START, 0, &(ctx->current));
    ctx->parent = ctx->pat->root;
    ctx->event = is_new ? PATRICIA_ADD_EVENT_NEW_NODE : PATRICIA_ADD_EVENT_NEXT_NODE;
    PATRICIA_ADD_HOOK;
  }

  while (ctx->word[ctx->word_pos]) {
    if ((ctx->current->follow_string != NULL) && (ctx->current->follow_string[ctx->follow_pos])) {
      /* There is a follow string and a the current follow string character is not null */

      if (ctx->word[ctx->word_pos] == ctx->current->follow_string[ctx->follow_pos]) {
	/* Current follow character and current word character are equal: proceed in follow string */

	ctx->word_pos++;
	ctx->follow_pos++;

	ctx->event = PATRICIA_ADD_EVENT_FOLLOW_STRING;
	PATRICIA_ADD_HOOK;
      }
      else {
	/* Current follow character and current word character are different: split the current node and its follow string */

	/* The result is two nodes: the first has the same leading character than the original one, the follow string is the same but stops just before the differing character, the second node has the differing character as leading character and the rest of the original follow string */
	ctx->parent = ctx->current;
	ctx->current = pnode_new(ctx->pat->nodes, ctx->parent, PNODE_TYPE_CHAR, ctx->current->follow_string[ctx->follow_pos]);
	if (ctx->parent->next != NULL)
	  ctx->parent->next->parent = ctx->current;/**/
	ctx->current->next = ctx->parent->next;
	ctx->parent->next = ctx->current;

	ctx->current->follow_string = ctx->parent->follow_string + ctx->follow_pos + 1;
	ctx->parent->follow_string[ctx->follow_pos] = L'\0';

	ctx->follow_pos = 0;
	ctx->event = PATRICIA_ADD_EVENT_SPLIT_FOLLOW_STRING;
	PATRICIA_ADD_HOOK;

	/* Add another node corresponding to the new word */
	pnode_insert(ctx->pat->nodes, ctx->parent, PNODE_TYPE_CHAR, ctx->word[ctx->word_pos], &(ctx->current));
	ctx->current->follow_string = bunch_claim(ctx->pat->follow_strings, wcslen(ctx->word) - ctx->word_pos);
	wcscat(ctx->current->follow_string, ctx->word + ctx->word_pos + 1);

	ctx->word_pos = wcslen(ctx->word);
	ctx->follow_pos = wcslen(ctx->current->follow_string);
	ctx->event = PATRICIA_ADD_EVENT_NEW_NODE;
	PATRICIA_ADD_HOOK;

	/* Add end of word node, if required */
	if (ctx->options & PATRICIA_ADD_OPTION_END) {
	  ctx->parent = ctx->current;
	  pnode_insert(ctx->pat->nodes, ctx->parent, PNODE_TYPE_END, 0, &(ctx->current));

	  ctx->follow_pos = 0;
	  ctx->event = PATRICIA_ADD_EVENT_NEW_NODE;
	  PATRICIA_ADD_HOOK;
	}

	return;
      }
    }
    else {
      /* No follow string or end of follow string: search/insert the next word character in children nodes */

      ctx->parent = ctx->current;
      if (pnode_insert(ctx->pat->nodes, ctx->parent, PNODE_TYPE_CHAR, ctx->word[ctx->word_pos], &(ctx->current))) {
	/* It is a new node: the rest of the word becomes the follow string */

	ctx->word_pos++;
	ctx->current->follow_string = bunch_claim(ctx->pat->follow_strings, wcslen(ctx->word) - ctx->word_pos + 1);
	wcscat(ctx->current->follow_string, ctx->word + ctx->word_pos);

	ctx->word_pos = wcslen(ctx->word);
	ctx->follow_pos = wcslen(ctx->current->follow_string);
	ctx->event = PATRICIA_ADD_EVENT_NEW_NODE;
	PATRICIA_ADD_HOOK;

	/* Add end of word node, if required */
	if (ctx->options & PATRICIA_ADD_OPTION_END) {
	  ctx->parent = ctx->current;
	  pnode_insert(ctx->pat->nodes, ctx->parent, PNODE_TYPE_END, 0, &(ctx->current));

	  ctx->follow_pos = 0;
	  ctx->event = PATRICIA_ADD_EVENT_NEW_NODE;
	  PATRICIA_ADD_HOOK;
	}

	return;
      }
      else {
	/* Node found: proceed */

	ctx->word_pos++;
	ctx->follow_pos = 0;

	ctx->event = PATRICIA_ADD_EVENT_NEXT_NODE;
	PATRICIA_ADD_HOOK;
      }
    }
  }

  /* Add end of word node, if required */ 
  if (ctx->options & PATRICIA_ADD_OPTION_END) {
    if ((ctx->current->follow_string != NULL) && (ctx->current->follow_string[ctx->follow_pos])) {
      /* Current follow character and current word character are different: split the current node and its follow string */

      /* The result is two nodes: the first has the same leading character than the original one, the follow string is the same but stops just before the differing character, the second node has the differing character as leading character and the rest of the original follow string */
      ctx->parent = ctx->current;
      ctx->current = pnode_new(ctx->pat->nodes, ctx->parent, PNODE_TYPE_CHAR, ctx->current->follow_string[ctx->follow_pos]);
      if (ctx->parent->next)
	ctx->parent->next->parent = ctx->current; /**/
      ctx->current->next = ctx->parent->next;
      ctx->parent->next = ctx->current;
      
      ctx->current->follow_string = ctx->parent->follow_string + ctx->follow_pos + 1;
      ctx->parent->follow_string[ctx->follow_pos] = 0;
 
      ctx->follow_pos = 0;
      ctx->event = PATRICIA_ADD_EVENT_SPLIT_FOLLOW_STRING;
      PATRICIA_ADD_HOOK;


      /* Add another node corresponding to the end of the new word */
      pnode_insert(ctx->pat->nodes, ctx->parent, PNODE_TYPE_END, 0, &(ctx->current));

      ctx->word_pos = wcslen(ctx->word);
      ctx->event = PATRICIA_ADD_EVENT_NEW_NODE;
      PATRICIA_ADD_HOOK;
    }
    else {
      /* No follow string or end of follow string: search/insert the end of word node in children nodes */

      ctx->parent = ctx->current;
      if (pnode_insert(ctx->pat->nodes, ctx->parent, PNODE_TYPE_END, 0, &(ctx->current)))
	ctx->event = PATRICIA_ADD_EVENT_NEW_NODE;
      else {
	ctx->event = PATRICIA_ADD_EVENT_NEXT_NODE;
      }

      ctx->follow_pos = 0;
      ctx->word_pos = wcslen(ctx->word);
      PATRICIA_ADD_HOOK;
    }
  }
}
#undef PATRICIA_ADD_HOOK


void patricia_compact_follow(patricia_t *pat) {
  bunch_t *new_follow_strings;
  unsigned int unclaimed = 0;
  unsigned int useless = 0;
  unsigned int i;
  unsigned int j;
  int was_zero = 1;
  
  for (i = 0; i < pat->follow_strings->block_number; i++) {
    unclaimed += pat->follow_strings->blocks[i]->unclaimed;
    for (j = 0; j < pat->follow_strings->block_size - pat->follow_strings->blocks[i]->unclaimed; j++)
      if (!((wchar_t *)pat->follow_strings->blocks[i]->data)[j]) {
	if (was_zero)
	  useless++;
	was_zero = 1;
      }
      else
	was_zero = 0;
  }

  new_follow_strings = bunch_new(sizeof(wchar_t), pat->follow_strings->block_number * pat->follow_strings->block_size - unclaimed - useless);
  for (i = 0; i < pat->nodes->block_number; i++)
    for (j = 0; j < pat->nodes->block_size - pat->nodes->blocks[i]->unclaimed; j++) {
      pnode_t *n = (pnode_t *)pat->nodes->blocks[i]->data + j;
     if (n->follow_string != NULL) {
	size_t l = wcslen(n->follow_string);
	if (l) {
	  wchar_t *s = bunch_claim(new_follow_strings, l + 1);
	  wcscat(s, n->follow_string);
	  n->follow_string = s;
	}
	else
	  n->follow_string = NULL;
      }
    }

  bunch_free(pat->follow_strings);
  pat->follow_strings = new_follow_strings;
}
