/*
  patricia.h - a PATRICIA tree structure and functions to create it, part of libtrish

  Copyright (C) 1988-1994,1996-1999,2003,2004 Free Software Foundation, Inc.
  This file is part of the GNU C Library.  Its master source is NOT part of
  the C library, however.  The master source lives in /gd/gnu/lib.

  The GNU C Library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  The GNU C Library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with the GNU C Library; if not, write to the Free
  Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
  02111-1307 USA.
*/

#ifndef HAVE_PATRICIA_H
#define HAVE_PATRICIA_H

#include "patricia_node.h"

/*
  patricia_t - PATRICIA tree structure
*/
typedef struct {
  /* nodes - bunch holding PATRICIA nodes */
  bunch_t *nodes;

  /* root - root of the tree */
  pnode_t *root;

  /* follow_strings - bunch holding follow strings */
  bunch_t *follow_strings;

  /* data - bunch holding data associated to the nodes */
  bunch_t *data;
} patricia_t;


/*
  patricia_new - creates a new PATRICIA tree, allocating it, creating nodes and follow strings bunches
  node_block - size of the nodes bunch blocks
  string_block - size of the follow strings blocks
  data - bunch holding data associated to the nodes, pass NULL if there is none
  return value - a freshly allocated PATRICIA tree

  Note: you are responsible for creating the data bunch beforehand
*/

patricia_t *patricia_new(unsigned int node_block, unsigned int string_block, bunch_t *data);


/*
  patricia_free - destroys a PATRICIA tree, frees nodes and follow strings bunches
  pat - tree to destroy

  Note: you are responsible for freeing the data bunch
*/

void patricia_free(patricia_t *pat);


/*
  patricia_save - writes a PATRICIA tree into a file
  pat - tree to write
  options - 0 or BUNCH_ATTRIB_PRUNED_BLOCKS to prune nodes, follow strings and data bunches
  f - file hande to write into, must be opened by fopen with mode "w" or "a"

  Note: this will write the data bunch
*/

void patricia_save(patricia_t *pat, int options, FILE *f);


/*
  patricia_load - reads a PATRICIA tree from a file
  f - file hande to read from, must be opened by fopen with mode "r"
  return value - a PATRICIA tree, no need to allocate it

  Note: patricia_load automatically correct node members next, left, right and data. If the data objects contain pointers to be corrected (pointers to nodes or to other data objects), you are responsible for it
 */

patricia_t *patricia_load(FILE *f);









/*
  patricia_add_context_t - a structure holding a context when adding a word to a PATRICIA tree
  pat - tree where the word is inserted
  options - OR'ed options (PNODE_TYPE_START to add a start of word node before the word, PNODE_TYPE_END to add a end of word node after the word)
  word - word to insert
  word_pos - index of the character in word currently added
  current - pointer to the node corresponding to the current character
  follow_pos - index of the character in the follow string of the current node
  event - type of the last addition event (see below)
  parent - pointer to the node parent to current
  arg - additional member to pass information to the hook (see below)
 */

typedef struct {
  patricia_t *pat;
  int options;
  const wchar_t *word;
  int word_pos;
  pnode_t *current;
  int follow_pos;
  int event;
  pnode_t *parent;
  void *arg;
} patricia_add_context_t;


/*
  PATRICIA_ADD_OPTION_* - addition options
 */

/* PATRICIA_ADD_OPTION_START - add start of word node at the beginning */
#define PATRICIA_ADD_OPTION_START 1

/* PATRICIA_ADD_OPTION_END - add end of word node at the end */
#define PATRICIA_ADD_OPTION_END   2


/*
  PATRICIA_ADD_EVENT_* - addition events
 */

/* PATRICIA_ADD_EVENT_NEW_NODE - current node has just been inserted */
#define PATRICIA_ADD_EVENT_NEW_NODE            1

/* PATRICIA_ADD_EVENT_SIDE_STRING_FORWARD - proceed in the follow string */
#define PATRICIA_ADD_EVENT_FOLLOW_STRING       2

/* PATRICIA_ADD_EVENT_NEXT_NODE - proceed at the end of the follow string to a child node */
#define PATRICIA_ADD_EVENT_NEXT_NODE           3

/* PATRICIA_ADD_EVENT_SPLIT_FOLLOW_STRING - parent follow string has been split */
#define PATRICIA_ADD_EVENT_SPLIT_FOLLOW_STRING 4


/*
  patricia_add_context_init - intitializes a context to add a new word
  ctx - a pointer to a context, you are responsible for allocating it
  pat - pointer to the PATRICIA tree where to add a new word
  word - word to add
  options - OR'ed options
  arg - additional info for the hook function

  Note: the pat, word, options and arg parameters will be passed as is to ctx, the other members are zeroed
 */

void patricia_add_context_init(patricia_add_context_t *ctx, patricia_t *pat, const wchar_t *word, int options, void *arg);


/*
  patricia_add - adds a word to a PATRICIA tree
  ctx - an adding context, must have been initialized with patricia_add_context_init
  hook - a function that will be called at each event, it will be passed the current context state, NULL means no hook

  Note: the hook is useful to set the user_info and data members of the nodes, modify other node members at your own risk, also modify the context members at your own risk 
 */

void patricia_add(patricia_add_context_t *ctx, void (*hook)(const patricia_add_context_t *));


/*
  patricia_compact_follow - compacts the follow strings bunch
  pat - tree to compact

  Note: the splitting of nodes in patricia_add produces unused consecutive L'\0' characters, this function replaces follow_strings by a new one where these are removed, it is useless to run it if the tree is not finalized yet.
*/

void patricia_compact_follow(patricia_t *pat);

#endif
