/*
  patricia_map.h - association of a string key to one or several void *, part of libtrish

  Copyright (C) 1988-1994,1996-1999,2003,2004 Free Software Foundation, Inc.
  This file is part of the GNU C Library.  Its master source is NOT part of
  the C library, however.  The master source lives in /gd/gnu/lib.

  The GNU C Library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  The GNU C Library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with the GNU C Library; if not, write to the Free
  Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
  02111-1307 USA.
*/

/*
  Summary:

  A simple map associates a string key to a single void * value. It is simply implemented as a patricia_t where values are claimed from it data member. The only functions provided here are pmap_add to add a new key/value pair and pmap_get to obtain the value from a key. To create, destroy, save and load, you may use the patricia_t functions. Just be sure that values are always claimed from its data member.

  A multi map associates a string key to a chain of void * values. It is implemented as a map to pmap_multi_t * (structure defined here). Values per se must be claimed from a single bunch you provide.
 */

#ifndef HAVE_PATRICIA_MAP_H
#define HAVE_PATRICIA_MAP_H

#include "patricia.h"


/*
  pmap_add - adds a key/value pair to a simple map
  pat - the map structure
  key - key
  value - value
  replace - if non-zero and the key already has a value, then replace the old value with the given one, otherwise let the existing pair unchanged
  return value - the old value if the key already had a value, NULL if the key didn't have a value

  Note: if value was not claimed from pat->data, then you will go to hell
*/

void *pmap_add(patricia_t *pat, wchar_t *key, void *value, int replace);


/*
  pmap_get - retrieves the value associated to a key
  pat - the map structure
  key - key to search for
  return value - the value associated to the key, NULL if none
*/

void *pmap_get(patricia_t *pat, wchar_t *key);


/*
  pmap_multival_t - a single chained list of values for multi maps
*/

typedef struct _pmap_multival_t pmap_multival_t;
struct _pmap_multival_t {
  /* value - the value */
  void *value;

  /* next - the next value, NULL if it is the last */
  pmap_multival_t *next;
};


/*
  pmap_multi_t - a multi map structure
*/

typedef struct {
  /* pat - keys */
  patricia_t *pat;

  /* values - values */
  bunch_t *values;
} pmap_multi_t;


/*
  pmap_multi_new - creates a new multi map
  node_block - block size for nodes for the keys tree
  string_block - block size for follow strings for the keys tree
  values - bunch where values come from
  return value - a freshly allocated multi map

  Note: you are the one who creates the values bunch
*/

pmap_multi_t *pmap_multi_new(unsigned int node_block, unsigned int string_block, bunch_t *values);


/*
  pmap_multi_free - destroys a multi map
  pmm - multi map to destroy

  Note: this function will not destroy the values bunch, you are responsible for it
*/

void pmap_multi_free(pmap_multi_t *pmm);


/*
  pmap_multi_add - adds a value to a key
  pmm - multi map
  key - key
  value - value to insert

  Note: if value was not claimed from pmm->values, you will suffer from a thousand deaths
*/

void pmap_multi_add(pmap_multi_t *pmm, wchar_t *key, void *value);


/*
  pmap_multi_get - retrieves the values associated to a key
  pmm - multi map structure
  key - key
  return value - the multival structure holding the first value associated to the key, NULL if none

  Note: the values are chained in reverse order of insertion, this may change in the future
*/

pmap_multival_t *pmap_multi_get(pmap_multi_t *pmm, wchar_t *key);


/*
  pmap_multi_save - writes a multi map to a file
  pmm - multi map structure to save
  options - same options as for bunch_save
  f - file handle to write into, must be opened by fopen with mode "w" or "a"

  Note: you cannot add key/values to a pruned multi map
*/

void pmap_multi_save(pmap_multi_t *pmm, int options, FILE *f);


/*
  pmap_multi_load - reads a multi map from a file
  f - file handle to read the multi map from, must be opened by fopen with mode "r"
  return value - the multi map
 */

pmap_multi_t *pmap_multi_load(FILE *f);

#endif
