/*
  patricia_node.h - a PATRICIA tree node structure, part of libtrish

  Copyright (C) 1988-1994,1996-1999,2003,2004 Free Software Foundation, Inc.
  This file is part of the GNU C Library.  Its master source is NOT part of
  the C library, however.  The master source lives in /gd/gnu/lib.

  The GNU C Library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  The GNU C Library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with the GNU C Library; if not, write to the Free
  Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
  02111-1307 USA.
*/

#ifndef HAVE_PATRICIA_NODE_H
#define HAVE_PATRICIA_NODE_H

#include "bunch.h"


/*
  pnode_t - structure holding a node
*/

typedef struct _pnode_t pnode_t;
struct _pnode_t {
  /* pn_type - PATRICIA node type */
  unsigned int pnode_type:3;

  /* avl_balance - unused yet, provisional for balancing siblings */
  unsigned int avl_balance:2;

  /* user_info - 3 reamining bits, use as you wish */
  unsigned int user_info:3;

  /* wc - leading character iff pn_type == PNODE_TYPE_CHAR */
  wchar_t wc;

  /* follow_string - following characters */
  wchar_t *follow_string;

  /* next - root of the tree of children nodes */
  pnode_t *next;

  /* left - root of the siblings with lower value */
  pnode_t *left;

  /* right - root of the siblings with greater value */
  pnode_t *right;

  /* parent - parent node, (node->next->parent == node) */
  pnode_t *parent;

  /* data - associated pointer, use as you wish */
  void *data;
};


/*
  PNODE_TYPE_* - PATRICIA node types
 */

/* PNODE_TYPE_ROOT - tree root */
#define PNODE_TYPE_ROOT      0

/* PNODE_TYPE_CHAR - character node */
#define PNODE_TYPE_CHAR      1

/* PNODE_TYPE_START - start of term node */
#define PNODE_TYPE_START     2

/* PNODE_TYPE_END - end of term node */
#define PNODE_TYPE_END       3


/*
  PNODE_AVL_* - balance values, unsupported yet
 */
#define PNODE_AVL_BALANCED 0
#define PNODE_AVL_LEFT     1
#define PNODE_AVL_RIGHT    3


/*
  pnode_new - claims a new node from a bunch
  b - bunch where to claim the new node
  pn_type - node type
  wc - node leading character
  return value - a pointer claimed from the bunch, NULL if the bunch has element_size different from sizeof(pnode_t) or if the bunch was pruned
 */

pnode_t *pnode_new(bunch_t *b, pnode_t *parent, int pn_type, wchar_t wc);


/*
  pnode_lookup - looks for a child of a node
  ref - node below which to look for
  pn_type - type of the searched node
  wc - leading character of the searched node
  return value - a pointer to the node child of ref with type pn_type and leading character wc, NULL if none was found
 */

pnode_t *pnode_lookup(pnode_t *ref, int pn_type, wchar_t wc);


/*
  pnode_insert - looks for a child of a node, inserts it if it has not been found
  b - node bunch where to claim a new node if necessary
  ref - node below which to look for
  pn_type - type of the searched node
  wc - leading character of the searched node
  new - a reference to a pointer to the searched node
  return value - 0 if the node has been found, 1 if pnode_insert had to insert it
 */

int pnode_insert(bunch_t *b, pnode_t *ref, int pn_type, wchar_t wc, pnode_t **new);


/*
  pnode_string - rebuilds the string corresponding to a node
  pn - the node
  n - length of the string
  return value - the reconstructed string

  Note: the string is claimed from bunch or allocated if NULL, also if n is too small the string will be truncated from the prefix, if n is too large the behaviour is undefined
*/

wchar_t *pnode_string(pnode_t *pn, int n, bunch_t *b);

#endif
