/*
  patricia_search.h - search words in a PATRICIA tree, part of libtrish

  Copyright (C) 1988-1994,1996-1999,2003,2004 Free Software Foundation, Inc.
  This file is part of the GNU C Library.  Its master source is NOT part of
  the C library, however.  The master source lives in /gd/gnu/lib.

  The GNU C Library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  The GNU C Library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with the GNU C Library; if not, write to the Free
  Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
  02111-1307 USA.
*/
#include "patricia_search.h"
#include "gci.h"

#include <wchar.h>
#include <wctype.h>

#include "mouchard.h"

psearch_t *psearch_new(patricia_t *pat, int options, patricia_t *gci, int density) {
  psearch_t *search = calloc(1, sizeof(psearch_t));

  search->pat = pat;
  search->options = options;
  search->gci = gci;
  search->solutions = bunch_new(sizeof(psearch_solution_t), density * 4 /* XXX shitty heuristic */);
  search->used = NULL;
  search->final = NULL;
  search->free = NULL;

  return search;
}

void psearch_reinit_stack(psearch_t *search, psearch_solution_t *stack) {
  for (search->current = stack; search->current; search->current = search->next) {
    search->next = search->current->next;
    search->current->query = NULL;
    search->current->query_pos = 0;
    search->current->node = NULL;
    search->current->follow_pos = 0;
    search->current->found_pos = 0;
    search->current->mismatch = 0;
    if (search->next == NULL) {
      search->current->next = search->free;
      if (search->free)
	search->free->prev = search->current;
    }
  }
  search->free = stack;
}

void psearch_reinit(psearch_t *search) {
  psearch_reinit_stack(search, search->used);
  psearch_reinit_stack(search, search->final);
  search->used = NULL;
  search->final = NULL;
  search->current = NULL;
  search->next = NULL;
}

void psearch_free(psearch_t *search) {
  bunch_free(search->solutions);
  free(search);
}

void psearch_new_query(psearch_t *search, wchar_t *query) {
  psearch_solution_t *sol;
  sol = psearch_solution_new(search, NULL, 0);
  sol->query = query;
  sol->node = search->pat->root;
}

psearch_solution_t *psearch_solution_new(psearch_t *search, psearch_solution_t *ref, int options) {
  psearch_solution_t *new;

  if (search->free == NULL) {
    new = bunch_claim(search->solutions, 1);
  }
  else {
    new = search->free;
    search->free = new->next;
  }
  if (ref == NULL) {
    new->next = search->used;
    new->prev = NULL;
    if (search->used != NULL)
      search->used->prev = new;
    search->used = new;
  }
  else {
    if (options & PSEARCH_SOLUTION_NEW_OPTION_BEFORE) {
      new->next = ref;
      new->prev = ref->prev;
      if (ref->prev == NULL)
	search->used = new;
      else
	ref->prev->next = new;
      ref->prev = new;
    }
    else {
      new->prev = ref;
      new->next = ref->next;
      if (ref->next != NULL)
	ref->next->prev = new;
      ref->next = new;
    }
    if (options & PSEARCH_SOLUTION_NEW_OPTION_CLONE) {
      new->query = ref->query;
      new->query_pos = ref->query_pos;
      new->node = ref->node;
      new->follow_pos = ref->follow_pos;
      new->found_pos = ref->found_pos;
      new->mismatch = ref->mismatch;
    }
  }

  return new;
}


void psearch_solution_finalize(psearch_t *search, psearch_solution_t *solution) {
  if (solution == search->used)
    search->used = solution->next;

  if (solution->next != NULL)
    solution->next->prev = solution->prev;
  if (solution->prev != NULL)
    solution->prev->next = solution->next;

  solution->prev = NULL;
  solution->next = search->final;
  if (search->final != NULL)
    search->final->prev = solution;

  search->final = solution;
}


void psearch_solution_discard(psearch_t *search, psearch_solution_t *solution) {
  if (solution == search->used)
    search->used = solution->next;

  if (solution->next != NULL)
    solution->next->prev = solution->prev;
  if (solution->prev != NULL)
    solution->prev->next = solution->next;

  solution->prev = NULL;
  solution->next = search->free;
  if (search->free != NULL)
    search->free->prev = solution;

  search->free = solution;

  solution->query = NULL;
  solution->query_pos = 0;
  solution->node = NULL;
  solution->follow_pos = 0;
  solution->found_pos = 0;
  solution->mismatch = 0;
}

void patricia_search(psearch_t *search) {
  while (search->used != NULL) {
    for (search->current = search->used; search->current; search->current = search->next) {
      wchar_t query_char = search->current->query[search->current->query_pos];
      wchar_t follow_char = (search->current->node->follow_string == NULL) ? 0 : search->current->node->follow_string[search->current->follow_pos];
      pnode_t *end_node = NULL;
      int matched_char = 0;

      search->next = search->current->next;

      /* match start of word */
      if ((search->current->node->pnode_type != PNODE_TYPE_START) && (search->options & PSEARCH_OPTION_START) && !search->current->query_pos) {
	pnode_t *start = pnode_lookup(search->current->node, PNODE_TYPE_START, 0);
	if (search->options & PSEARCH_OPTION_START_MANDATORY) {
	  if (start == NULL) {
	    psearch_solution_discard(search, search->current);
	    break;
	  }
	  else
	    search->current->node = start;
	}
	else {
	  if (start != NULL) {
	    psearch_solution_t *next = psearch_solution_new(search, search->current, PSEARCH_SOLUTION_NEW_OPTION_CLONE);
	    next->node = start;
	    search->next = next;
	  }
	}
      }

      /* match end of word */
      if (!follow_char && (search->options && PSEARCH_OPTION_END) && !((search->options & PSEARCH_OPTION_WHOLE) && query_char)) {
	if ((end_node = pnode_lookup(search->current->node, PNODE_TYPE_END, 0))) {
	  psearch_solution_t *sol = psearch_solution_new(search, search->current, PSEARCH_SOLUTION_NEW_OPTION_CLONE);
	  sol->node = end_node;
       	  sol->follow_pos = 0;
	  psearch_solution_finalize(search, sol);
	}
      }

      /* case insensitive match */
      if (query_char && (search->options & (search->current->query_pos ? PSEARCH_OPTION_CI_TAIL : PSEARCH_OPTION_CI_FIRST)) && iswalpha(query_char)) {
	wchar_t case_query_char = iswlower(query_char) ? towupper(query_char) : towlower(query_char);
	if (follow_char) {
	  if (case_query_char == follow_char) {
	    psearch_solution_t *sol = psearch_solution_new(search, search->current, PSEARCH_SOLUTION_NEW_OPTION_BEFORE | PSEARCH_SOLUTION_NEW_OPTION_CLONE);
	    sol->follow_pos++;
	    sol->query_pos++;
	    sol->found_pos++;
	    sol->mismatch++;
	  }
	}
	else {
	  pnode_t *char_node;
	  if ((char_node = pnode_lookup(search->current->node, PNODE_TYPE_CHAR, case_query_char))) {
	    psearch_solution_t *sol = psearch_solution_new(search, search->current, PSEARCH_SOLUTION_NEW_OPTION_BEFORE | PSEARCH_SOLUTION_NEW_OPTION_CLONE);
	    sol->query_pos++;
	    sol->follow_pos = 0;
	    sol->node = char_node;
	    sol->found_pos++;
	    sol->mismatch++;
	  }
	}
     }

      /* GCI matches */
      if ((search->gci != NULL) && (search->options & (search->current->query_pos ? PSEARCH_OPTION_GCI_TAIL : PSEARCH_OPTION_GCI_FIRST))) {
	wchar_t *query_alt = gci_alt(search->gci, query_char);
	if (follow_char) {
	  if (query_char && (wcschr(query_alt, follow_char))) {
	    matched_char = 1;
	    search->current->follow_pos++;
	    search->current->query_pos++;
	    search->current->found_pos++;
	    search->current->mismatch++;
	  }
	  else if ((search->options & (query_char ? PSEARCH_OPTION_CI_TAIL : PSEARCH_OPTION_CI_FIRST)) && iswalpha(query_char)) {
	  }
	}
	else {
	  int i;
	  for (i = 0; query_alt[i]; i++) {
	    pnode_t *char_node;
	    if ((char_node = pnode_lookup(search->current->node, PNODE_TYPE_CHAR, query_char))) {
	      psearch_solution_t *sol = psearch_solution_new(search, search->current, PSEARCH_SOLUTION_NEW_OPTION_BEFORE | PSEARCH_SOLUTION_NEW_OPTION_CLONE);
	      sol->query_pos++;
	      sol->follow_pos = 0;
	      sol->node = char_node;
	      sol->found_pos++;
	      sol->mismatch++;
	    }
	  }
	}
      }

      /* exact match */
      if (follow_char) {
	if (query_char && (query_char == follow_char)) {
	  matched_char = 1;
	  search->current->follow_pos++;
	  search->current->query_pos++;
	  search->current->found_pos++;
	}
      }
      else {
	pnode_t *char_node;
	if (query_char && (char_node = pnode_lookup(search->current->node, PNODE_TYPE_CHAR, query_char))) {
	  matched_char = 1;
	  search->current->query_pos++;
	  search->current->follow_pos = 0;
	  search->current->node = char_node;
	  search->current->found_pos++;
	}
      }

      /* in case no character has been matched */
      if (!matched_char) {
	if (((search->options & PSEARCH_OPTION_WHOLE) && query_char) || (search->options & PSEARCH_OPTION_END_MANDATORY) || end_node) {
	  psearch_solution_discard(search, search->current);
	}
	else {
	  psearch_solution_finalize(search, search->current);
	}
      }
    }
  }
}
  
