/*
  patricia_search.h - search words in a PATRICIA tree, part of libtrish

  Copyright (C) 1988-1994,1996-1999,2003,2004 Free Software Foundation, Inc.
  This file is part of the GNU C Library.  Its master source is NOT part of
  the C library, however.  The master source lives in /gd/gnu/lib.

  The GNU C Library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  The GNU C Library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with the GNU C Library; if not, write to the Free
  Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
  02111-1307 USA.
*/

#ifndef HAVE_PATRICIA_SEARCH_H
#define HAVE_PATRICIA_SEARCH_H

#include "patricia.h"


/*
  psearch_solution_t - a structure holding a single solution
 */
typedef struct _psearch_solution_t psearch_solution_t;
struct _psearch_solution_t {
  /* query - the query word */
  wchar_t *query;

  /* query_pos - index of the currently searched character */
  int query_pos;

  /* node - current node */
  pnode_t *node;

  /* follow_pos - index of the current character in the follow string */
  int follow_pos;

  /* found_pos - length of the found string */
  size_t found_pos;

  /* mismatch - number of mismatches */
  int mismatch;

  /* next/prev - solutions in double link list */
  psearch_solution_t *next;
  psearch_solution_t *prev;
};


/*
  psearch_t - structure holding a search context
*/

typedef struct {
  /* pat - PATRICIA tree where to search */
  patricia_t *pat;

  /* options - OR'ed options */
  int options;

  /* gci - general case insensitive table */
  patricia_t *gci;

  /* solutions - bunch of psearch_solution_t */
  bunch_t *solutions;

  /* current - used in psearch iteration, the solution currently examined */
  psearch_solution_t *current;

  /* next - used in psearch iteration, the solution examined after current */
  psearch_solution_t *next;

  /* used - pointer to the first currently examined solution */
  psearch_solution_t *used;

  /* used - pointer to the first finalized solution */
  psearch_solution_t *final;

  /* free - pointer to the first discarded solution */
  psearch_solution_t *free;
} psearch_t;


/*
  PSEARCH_OPTION_* - options for patricia_search
 */

/* PSEARCH_OPTION_START - may match start of word at the beginning */
#define PSEARCH_OPTION_START           1

/* PSEARCH_OPTION_START_MANDATORY - must match start of word at the beginning */
#define PSEARCH_OPTION_START_MANDATORY (2 | PSEARCH_OPTION_START)

/* PSEARCH_OPTION_WHOLE - must match all characters until the end of the query */
#define PSEARCH_OPTION_WHOLE           4

/* PSEARCH_OPTION_END - may match end of word at any time */
#define PSEARCH_OPTION_END             8

/* PSEARCH_OPTION_END_MANDATORY - must match end of word */
#define PSEARCH_OPTION_END_MANDATORY   (16 | PSEARCH_OPTION_END)

/* PSEARCH_OPTION_FULL - all options on */
#define PSEARCH_OPTION_FULL            (PSEARCH_OPTION_START_MANDATORY | PSEARCH_OPTION_WHOLE | PSEARCH_OPTION_END_MANDATORY)

/* PSEARCH_OPTION_TAG - must match begin and end of word nodes */
#define PSEARCH_OPTION_TAG             (PSEARCH_OPTION_START_MANDATORY | PSEARCH_OPTION_END_MANDATORY)

/* PSEARCH_OPTION_CI_FIRST - first query character must be matched case insensitively */
#define PSEARCH_OPTION_CI_FIRST        32

/* PSEARCH_OPTION_CI_TAIL - query characters but first must be matched case insensitively */
#define PSEARCH_OPTION_CI_TAIL         64

/* PSEARCH_OPTION_CI_ALL - all query characters must be matched case insensitively */
#define PSEARCH_OPTION_CI_ALL          (PSEARCH_OPTION_CI_FIRST | PSEARCH_OPTION_CI_TAIL)

/* PSEARCH_OPTION_GCI_FIRST - first query character must be matched with GCI table, ignored if gci is NULL */
#define PSEARCH_OPTION_GCI_FIRST       128

/* PSEARCH_OPTION_GCI_TAIL - query characters but first must be matched with GCI table, ignored if gci is NULL */
#define PSEARCH_OPTION_GCI_TAIL        256

/* PSEARCH_OPTION_GCI_ALL - all query characters must be matched with GCI table, ignored if gci is NULL */
#define PSEARCH_OPTION_GCI_ALL         (PSEARCH_OPTION_GCI_FIRST | PSEARCH_OPTION_GCI_TAIL)


/*
  psearch_new - creates a new search
  pat - tree to search in
  options - OR'ed PSEARCH_OPTION_*
  gci - general case insensitive table
  density - estimate of the number of solutions, better overestimate than underestimate
  return value - a fresh pointer to a search structure
*/

psearch_t *psearch_new(patricia_t *pat, int options, patricia_t *gci, int density);


/*
  psearch_reinit - reinitializes a search structure but keeps options and patricia
  search - structure to reinitialize
 */

void psearch_reinit(psearch_t *search);


/*
  psearch_free - destroys a search and all its solutions
  search - structure to destroy

  Note: the PATRICIA tree won't be freed
 */

void psearch_free(psearch_t *search);


/*
  psearch_new_query - adds a new query string
  search - search structure where to add the query
  query - string to search for

  Note: query will not be copied, so do not modify or free it in the middle of a search operation...
*/

void psearch_new_query(psearch_t *search, wchar_t *query);


/*
  psearch_solution_new - adds a new solution to a search
  search - a search structure as given by psearch_new
  ref - a reference solution
  options - OR'ed PSEARCH_SOLUTION_NEW_OPTION_*
  return value - a pointer to a fresh solution

  Note: the new solution will be part of the linked list of solutions kept by the search structure, ref must be a solution of the same search structure, if not the behaviour is undefined
*/

psearch_solution_t *psearch_solution_new(psearch_t *search, psearch_solution_t *ref, int options);


/*
  PSEARCH_SOLUTION_NEW_OPTION_* - options for psearch_solution_new
*/

/* PSEARCH_SOLUTION_NEW_OPTION_BEFORE - inserts the new solution before the reference solution, by default it inserts after */
#define PSEARCH_SOLUTION_NEW_OPTION_BEFORE 1

/* PSEARCH_SOLUTION_NEW_OPTION_CLONE - the new solution will be a copy of the reference solution, by default all members are set to 0/NULL */
#define PSEARCH_SOLUTION_NEW_OPTION_CLONE  2


/*
  psearch_solution_finalize - registers a solution from the search as final
  search - the search structure the solution belongs
  solution - solution to finalize

  Note: if solution was not given by search, the behaviour is undefined
 */

void psearch_solution_finalize(psearch_t *search, psearch_solution_t *solution);


/*
  psearch_solution_discard - removes a solution from the search
  search - the search structure the solution belongs
  solution - solution to remove

  Note: if solution was not given by search, the behaviour is undefined
 */

void psearch_solution_discard(psearch_t *search, psearch_solution_t *solution);


/*
  psearch - start a search
  search - the search to proceed

  Note: when this function returns, all solutions are held in search->final
 */

void patricia_search(psearch_t *search);


#endif
