/*
  string_collection.h - manages a collection of strings, part of libtrish

  Copyright (C) 1988-1994,1996-1999,2003,2004 Free Software Foundation, Inc.
  This file is part of the GNU C Library.  Its master source is NOT part of
  the C library, however.  The master source lives in /gd/gnu/lib.

  The GNU C Library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  The GNU C Library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with the GNU C Library; if not, write to the Free
  Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
  02111-1307 USA.
*/

/*
  Summary:

  A string collection is a set of pointers to strings that guarantees that each different string is only stored once. The PATRICIA structure allows to search for existing strings in the collection. Its data member is a bunch of wchar_t containing all the strings of the collection.

*/

#ifndef HAVE_STRING_COLLECTION_H
#define HAVE_STRING_COLLECTION_H

#include "patricia.h"


/*
  string_collection_new - creates a new string collection
  node_block - node block sizee for the tree
  string_block - string block size for the tree
  collection_size - estimated size of the string collection in characters
  return value - a string collection PATRICIA tree
*/

patricia_t *string_collection_new(unsigned int node_block, unsigned int string_block, unsigned int collection_size);


/*
  string_collection_free - destroys a string collection and keeps the strings
  scoll - the string collection to destroy
  return value - the data member, thus all strings in the collection
*/

bunch_t *string_collection_free(patricia_t *scoll);


/*
  string_collection_save - saves a string collection into a file
  scoll - string collection to save
  f - file handle to save into, must be opened by fopen with mode "w" or "a"

  Note: this is an alias for patricia_save(scoll, 0, f) but it is implemented as a function
*/

void string_collection_save(patricia_t *scoll, FILE *f);


/*
  string_collection_load - loads a string collection from a file
  f - file handle to read from, must be opened by fopen with mode "r"
  return value - a string collection

  Note: this is an alias for patricia_load(f) but it is implemented as a function
 */

patricia_t *string_collection_load(FILE *f);


/*
  string_collection_ptr - searches and inserts if necessary a string into a collection and returns a pointer to a copy of this string in the data member
  scoll - string collection to look into
  w - string to look for
  return value - a claimed string

  Note: w != return value, string is copied so you may free w afterwards
*/

wchar_t *string_collection_ptr(patricia_t *scoll, const wchar_t *w);

#endif
