#include "defaults.h"
#include "cl_config.h"
#include "verbosity.h"
#include "patricia_search.h"
#include "wchar_utils.h"
#include "trish_dict.h"
#include "gci.h"

#include <popt.h>
#include <locale.h>
#include <wchar.h>
#include <errno.h>
#include <wctype.h>

#include "mouchard.h"

FILE *open_file(const char *fn, int write) {
  FILE *fh = fopen(fn, write ? "w" : "r");

  if (fh == NULL)
    verbosity_message(V_QUIET | V_ERROR, "Could not open file '%s'", fn);

  return fh;
}

void set_encoding(char *encoding) {
  if (setlocale(LC_CTYPE, encoding) == NULL) {
    verbosity_message(V_QUIET | V_ERROR, "Unhandled locale '%s'", encoding);
  }
  verbosity_message(V_NORMAL | V_INFO, "Locale set to '%s'", encoding);
}

int split_line(wchar_t *line, const wchar_t *sep, wchar_t **split) {
  size_t i = 0;

  if (sep == NULL)
    return 1;

  while (*line) {
    if (*line == sep[i]) {
      *line = L'\0';
      line++;
      split[i] = line;
      i++;
    }
    else
      line++;
  }

  return !sep[i];
}



typedef struct _display_t display_t;
struct _display_t {
  int op;
  union {
    wchar_t *str;
    display_t *loop;
    intptr_t tag;
  } arg;
  display_t *next;
};

#define DISPLAY_OP_NULL     0
#define DISPLAY_OP_STRING   1
#define DISPLAY_OP_QUERY    2
#define DISPLAY_OP_LINE     3
#define DISPLAY_OP_START    4
#define DISPLAY_OP_LENGTH   5
#define DISPLAY_OP_MATCH    6
#define DISPLAY_OP_ENTRY    7
#define DISPLAY_OP_MISMATCH 8
#define DISPLAY_OP_SOM      9
#define DISPLAY_OP_EOM      10
#define DISPLAY_OP_TAG      11

#define DISPLAY_OPEN_WCHAR  L'{'
#define DISPLAY_CLOSE_WCHAR L'}'
#define DISPLAY_SOM_WCHAR   L'['
#define DISPLAY_EOM_WCHAR   L']'

display_t *display_add(bunch_t *b, display_t *prev, int op, void *arg) {
  display_t *d = bunch_claim(b, 1);
  d->op = op;
  switch (op) {
  case DISPLAY_OP_STRING:
    d->arg.str = arg;
    break;
  case DISPLAY_OP_EOM:
    d->arg.loop = arg;
    break;
  case DISPLAY_OP_TAG:
    d->arg.tag = (intptr_t)arg;
    break;
  }
  if (prev != NULL)
    prev->next = d;
  return d;
}

display_t *format_display(bunch_t *b, wchar_t *fmt, int yes, int multi, int n_tags) {
  display_t *first = display_add(b, NULL, DISPLAY_OP_NULL, NULL);
  display_t *last = first;
  display_t *loop = NULL;
  int i;
  wchar_t *op = NULL;
  wchar_t *str = fmt;
  wchar_t *bstr = NULL;
  int err = 0;

#define NEW_STR if (bstr != NULL) {\
  *str = L'\0';\
  str++;\
  last = display_add(b, last, DISPLAY_OP_STRING, bstr);\
  bstr = NULL;\
}
#define SOERR {\
  err++;\
  verbosity_message(V_NORMAL | V_INFO | V_NONL, "Format %s, char %i : ", yes ? "yes" : "no", i);\
  verbosity_message(V_NORMAL | V_WARNING | V_NOHDR,
#define EOERR ); }

  if (fmt == NULL)
    return NULL;

  for (i = 0; fmt[i]; i++) {
    if (op != NULL) {
      if (fmt[i] == DISPLAY_CLOSE_WCHAR) {
	fmt[i] = L'\0';
	if (!wcsncmp(op, L"query", 5)) {
	  last = display_add(b, last, DISPLAY_OP_QUERY, NULL);
	}
	else if (!yes) {
	  SOERR "Illegal '%ls'", op EOERR;
	}
	else {
	  if (!wcsncmp(op, L"line", 4)) {
	    last = display_add(b, last, DISPLAY_OP_LINE, NULL);
	  }
	  else if (!wcsncmp(op, L"start", 5)) {
	    last = display_add(b, last, DISPLAY_OP_START, NULL);
	  }
	  else if (!wcsncmp(op, L"length", 6)) {
	    last = display_add(b, last, DISPLAY_OP_LENGTH, NULL);
	  }
	  else if (!wcsncmp(op, L"match", 5)) {
	    last = display_add(b, last, DISPLAY_OP_MATCH, NULL);
	  }
	  else if (!wcsncmp(op, L"entry", 5)) {
	    last = display_add(b, last, DISPLAY_OP_ENTRY, NULL);
	  }
	  else if (!wcsncmp(op, L"mismatch", 8)) {
	    last = display_add(b, last, DISPLAY_OP_MISMATCH, NULL);
	  }
	  else if (wcsstr(op, L"tag ") == op) {
	    if (multi && (loop == NULL)) {
	      SOERR "In multi dictionaries, tag display must be between multi paren" EOERR;
	    }
	    else {
	      intptr_t tag;
	      op += 4;
	      tag = *op - L'0';
	      op++;
	      if (*op != L'\0') {
		op -= 5;
		SOERR "Unknown '%ls'", op EOERR;
	      }
	      else if ((tag < 0) || (tag >= n_tags)) {
		SOERR "Illegal tag number '%d'", tag EOERR;
	      }
	      else {
		last = display_add(b, last, DISPLAY_OP_TAG, (void *)tag);
	      }
	    }
	  }
	  else {
	    SOERR "Unknown '%ls'", op EOERR;
	  }
	}
	op = NULL;
      }
    }
    else {
      switch (fmt[i]) {
      case DISPLAY_OPEN_WCHAR:
	NEW_STR;
	op = fmt + i + 1;
	break;
      case DISPLAY_SOM_WCHAR:
	NEW_STR;
	if (!yes) {
	  SOERR "Multi paren not allowed in format for unmatched queries" EOERR;
	  break;
	}
	if (!multi) {
	  SOERR "This dictionary is not multi" EOERR;
	  break;
	}
	if (loop != NULL) {
	  SOERR "double SOM" EOERR;
	  break;
	}
	last = display_add(b, last, DISPLAY_OP_SOM, NULL);
	loop = last;
	break;
      case DISPLAY_EOM_WCHAR:
	NEW_STR;
	if (!yes) {
	  SOERR "Multi paren not allowed in format for unmatched queries" EOERR;
	  break;
	}
	if (!multi) {
	  SOERR "This dictionary is not multi" EOERR;
	  break;
	}
	if (loop == NULL) {
	  SOERR "mismatch EOM" EOERR;
	  break;
	}
	last = display_add(b, last, DISPLAY_OP_EOM, loop);
	loop = NULL;
	break;
      default:
	if (bstr == NULL)
	  bstr = str;
	*str = fmt[i];
	str++;
	break;
      }
    }
  }
  NEW_STR;

  if (op != NULL) {
    SOERR "Unmatched '{'" EOERR;
  }

  if (loop != NULL) {
    SOERR "Unmatched '%lc'", DISPLAY_SOM_WCHAR EOERR;
  }

  if (err) {
    verbosity_message(VERBOSITY_ERROR, "There are %d errors in the %s format", err, yes ? "yes" : "no");
  }

  return first;

#undef NEW_STR
#undef SOERR
#undef EOERR
}

void display_headers(FILE *f, tdict_t *td, display_t *display) {
  if (display == NULL)
    return;

  if (td->entry_header == NULL)
    return;

  while (display != NULL) {
    switch (display->op) {
    case DISPLAY_OP_NULL:
      break;
    case DISPLAY_OP_STRING:
      fprintf(f, "%ls", display->arg.str);
      break;
    case DISPLAY_OP_QUERY:
      fprintf(f, "query");
      break;
    case DISPLAY_OP_LINE:
      fprintf(f, "line");
      break;
    case DISPLAY_OP_START:
      fprintf(f, "start");
      break;
    case DISPLAY_OP_LENGTH:
      fprintf(f, "length");
      break;
    case DISPLAY_OP_MATCH:
      fprintf(f, "match");
      break;
    case DISPLAY_OP_ENTRY:
      fprintf(f, "%ls",  td->entry_header);
      break;
    case DISPLAY_OP_MISMATCH:
      fprintf(f, "mismatch");
      break;
    case DISPLAY_OP_SOM:
      break;
    case DISPLAY_OP_EOM:
      break;
    case DISPLAY_OP_TAG:
      fprintf(f, "%ls",  td->tag_headers[display->arg.tag]);
      break;
    }
    display = display->next;
  }
}

void display(FILE *f, tdict_t *td, wchar_t *query, unsigned int line_nb, psearch_solution_t *sol, display_t *display) {
  pmap_multival_t *mval = NULL;
  wchar_t *ws;
  wchar_t wc;

  if (display == NULL)
    return;

  while (display != NULL) {
    switch (display->op) {
    case DISPLAY_OP_NULL:
      break;
    case DISPLAY_OP_STRING:
      fprintf(f, "%ls", display->arg.str);
      break;
    case DISPLAY_OP_QUERY:
      fprintf(f, "%ls", query);
      break;
    case DISPLAY_OP_LINE:
      fprintf(f, "%ud", line_nb);
      break;
    case DISPLAY_OP_START:
      fprintf(f, "%d", (int)((intptr_t)sol->query - (intptr_t)query));
      break;
    case DISPLAY_OP_LENGTH:
      fprintf(f, "%d", sol->query_pos);
      break;
    case DISPLAY_OP_MATCH:
      wc = sol->query[sol->query_pos];
      sol->query[sol->query_pos] = L'\0';
      fprintf(f, "%ls", sol->query);
      sol->query[sol->query_pos] = wc;
      break;
    case DISPLAY_OP_ENTRY:
      ws = pnode_string(sol->node, sol->found_pos, NULL);
      fprintf(f, "%ls", ws);
      free(ws);
      break;
    case DISPLAY_OP_MISMATCH:
      fprintf(f, "%d", sol->mismatch);
      break;
    case DISPLAY_OP_SOM:
      mval = sol->node->data;
      break;
    case DISPLAY_OP_EOM:
      mval = mval->next;
      if (mval != NULL)
	display = display->arg.loop;
      break;
    case DISPLAY_OP_TAG:
      fprintf(f, "%ls", (mval == NULL) ? ((wchar_t **)sol->node->data)[display->arg.tag] : ((wchar_t **)mval->value)[display->arg.tag]);
      break;
    }
    display = display->next;
  }
}

int main(int argc, const char **argv) {

  static unsigned int optimize_params[6] = { DEFAULT_OPT_MNB, DEFAULT_OPT_MSB, DEFAULT_OPT_TNB, DEFAULT_OPT_TSB, DEFAULT_OPT_NENTRIES, DEFAULT_OPT_TAGSIZE };
  static int compiled_dict = 0;
  static size_t dict_max_length = DEFAULT_MAX_LINE_LENGTH;
  static int multi_dict = 0;
  static char *dict_encoding = DEFAULT_ENCODING;
  static int headers = 0;
  static char *in_fn = NULL;
  static char *out_fn = NULL;
  static size_t search_max_length = DEFAULT_MAX_LINE_LENGTH;
  static char *search_encoding = DEFAULT_ENCODING;
  static int substrings = 0;
  static int words = 0;
  static char *gci_fn = NULL;
  static int show_headers = 0;
  static int no_search = 0;
  static char *save_as = NULL;
  static int display_opt = 0;
  int search_options = DEFAULT_SEARCH_OPTIONS;
  static char *dummy_format_yes = DEFAULT_FORMAT_YES;
  wchar_t *format_yes = char2wcs(DEFAULT_FORMAT_YES);
  wchar_t *format_no = NULL;
  const char *dict_fn = NULL;

  static struct poptOption cl_verbosity_options[] = CL_VERBOSITY_OPTIONS;

  static struct poptOption cl_opt_options[] = {
    { CLOL_OPT_MNB        , CLOS_OPT_MNB        , POPT_ARG_INT    | POPT_ARGFLAG_SHOW_DEFAULT, &optimize_params[0] , 0                         , CLOH_OPT_MNB        , "INT" },
    { CLOL_OPT_MSB        , CLOS_OPT_MSB        , POPT_ARG_INT    | POPT_ARGFLAG_SHOW_DEFAULT, &optimize_params[1] , 0                         , CLOH_OPT_MSB        , "INT" },
    { CLOL_OPT_TNB        , CLOS_OPT_TNB        , POPT_ARG_INT    | POPT_ARGFLAG_SHOW_DEFAULT, &optimize_params[2] , 0                         , CLOH_OPT_TNB        , "INT" },
    { CLOL_OPT_TSB        , CLOS_OPT_TSB        , POPT_ARG_INT    | POPT_ARGFLAG_SHOW_DEFAULT, &optimize_params[3] , 0                         , CLOH_OPT_TSB        , "INT" },
    { CLOL_OPT_NENTRIES   , CLOS_OPT_NENTRIES   , POPT_ARG_INT    | POPT_ARGFLAG_SHOW_DEFAULT, &optimize_params[4] , 0                         , CLOH_OPT_NENTRIES   , "INT" },
    { CLOL_OPT_TAGSIZE    , CLOS_OPT_TAGSIZE    , POPT_ARG_INT    | POPT_ARGFLAG_SHOW_DEFAULT, &optimize_params[5] , 0                         , CLOH_OPT_TAGSIZE    , "INT" },
    POPT_TABLEEND
  };

  static struct poptOption cl_dict_options[] = {
    { CLOL_COMPILED_DICT   , CLOS_COMPILED_DICT   , POPT_ARG_NONE                              , &compiled_dict      , 0                   , CLOH_COMPILED_DICT   , NULL },
    { CLOL_DICT_MAX_LENGTH , CLOS_DICT_MAX_LENGTH , POPT_ARG_INT    | POPT_ARGFLAG_SHOW_DEFAULT, &dict_max_length    , 0                   , CLOH_DICT_MAX_LENGTH , "LENGTH" },
    { CLOL_TAG_DELIMITERS  , CLOS_TAG_DELIMITERS  , POPT_ARG_STRING                            , NULL                , CLOV_TAG_DELIMITERS , CLOH_TAG_DELIMITERS  , "DELIMITERS" },
    { CLOL_MULTI_DICT      , CLOS_MULTI_DICT      , POPT_ARG_NONE                              , &multi_dict         , 0                   , CLOH_MULTI_DICT      , NULL },
    { CLOL_DICT_ENCODING   , CLOS_DICT_ENCODING   , POPT_ARG_STRING | POPT_ARGFLAG_SHOW_DEFAULT, &dict_encoding      , 0                   , CLOH_DICT_ENCODING   , "ENCODING" },
    { CLOL_HEADERS         , CLOS_HEADERS         , POPT_ARG_NONE                              , &headers            , 0                   , CLOH_HEADERS         , NULL },
    POPT_TABLEEND
  };

  static struct poptOption cl_search_options[] = {
    { CLOL_INPUT             , CLOS_INPUT             , POPT_ARG_STRING                            , &in_fn              , 0                   , CLOH_INPUT             , "FILE" },
    { CLOL_OUTPUT            , CLOS_OUTPUT            , POPT_ARG_STRING                            , &out_fn             , 0                   , CLOH_OUTPUT            , "FILE" },
    { CLOL_SEARCH_MAX_LENGTH , CLOS_SEARCH_MAX_LENGTH , POPT_ARG_INT    | POPT_ARGFLAG_SHOW_DEFAULT, &search_max_length  , 0                   , CLOH_SEARCH_MAX_LENGTH , "LENGTH" },
    { CLOL_SEARCH_ENCODING   , CLOS_SEARCH_ENCODING   , POPT_ARG_STRING | POPT_ARGFLAG_SHOW_DEFAULT, &search_encoding    , 0                   , CLOH_SEARCH_ENCODING   , "ENCODING" },
    { CLOL_SUBSTRINGS        , CLOS_SUBSTRINGS        , POPT_ARG_NONE                              , &substrings         , CLOV_SUBSTRINGS     , CLOH_SUBSTRINGS        , NULL },
    { CLOL_WORDS             , CLOS_WORDS             , POPT_ARG_NONE                              , &words              , 0                   , CLOH_WORDS             , NULL },
    { CLOL_CI_FIRST          , CLOS_CI_FIRST          , POPT_ARG_NONE                              , NULL                , CLOV_CI_FIRST       , CLOH_CI_FIRST          , NULL },
    { CLOL_CI_ALL            , CLOS_CI_ALL            , POPT_ARG_NONE                              , NULL                , CLOV_CI_ALL         , CLOH_CI_ALL            , NULL },
    { CLOL_GCI_FIRST         , CLOS_GCI_FIRST         , POPT_ARG_NONE                              , NULL                , CLOV_GCI_FIRST      , CLOH_GCI_FIRST         , NULL },
    { CLOL_GCI_ALL           , CLOS_GCI_ALL           , POPT_ARG_NONE                              , NULL                , CLOV_GCI_ALL        , CLOH_GCI_ALL           , NULL },
    { CLOL_GCI_TABLE         , CLOS_GCI_TABLE         , POPT_ARG_STRING                            , &gci_fn             , 0                   , CLOH_GCI_TABLE         , "PATH" },
    { CLOL_FORMAT_YES        , CLOS_FORMAT_YES        , POPT_ARG_STRING | POPT_ARGFLAG_SHOW_DEFAULT, &dummy_format_yes   , CLOV_FORMAT_YES     , CLOH_FORMAT_YES        , "FORMAT" },
    { CLOL_FORMAT_NO         , CLOS_FORMAT_NO         , POPT_ARG_STRING                            , NULL                , CLOV_FORMAT_NO      , CLOH_FORMAT_NO         , "FORMAT" },
    { CLOL_SHOW_HEADERS      , CLOS_SHOW_HEADERS      , POPT_ARG_NONE                              , &show_headers       , 0                   , CLOH_SHOW_HEADERS      , NULL },
    POPT_TABLEEND
  };

  static struct poptOption cl_options[] = {
    POPT_AUTOHELP
    { CLOL_NO_SEARCH         , CLOS_NO_SEARCH       , POPT_ARG_NONE   , &no_search          , 0                   , CLOH_NO_SEARCH       , NULL },
    { CLOL_SAVE_AS           , CLOS_SAVE_AS         , POPT_ARG_STRING , &save_as            , 0                   , CLOH_SAVE_AS         , "FILE" },
    { CLOL_DISPLAY_OPT       , CLOS_DISPLAY_OPT     , POPT_ARG_NONE   , &display_opt        , 0                   , CLOH_DISPLAY_OPT     , NULL },
    { CLOL_VERSION           , CLOS_VERSION         , POPT_ARG_NONE   , NULL                , CLOV_VERSION        , CLOH_VERSION         , NULL },
    { NULL , '\0' , POPT_ARG_INCLUDE_TABLE , cl_verbosity_options , 0 , "Verbosity options:" , NULL },
    { NULL , '\0' , POPT_ARG_INCLUDE_TABLE , cl_search_options    , 0 , "Search options:" , NULL },
    { NULL , '\0' , POPT_ARG_INCLUDE_TABLE , cl_dict_options , 0 , "Dictionary options:" , NULL },
    { NULL , '\0' , POPT_ARG_INCLUDE_TABLE , cl_opt_options , 0 , "Obscure optimization options:" , NULL },
    POPT_TABLEEND
  };

  poptContext cl_context = poptGetContext(NULL, argc, argv, cl_options, 0);
  int cl_oh;
  FILE *dict_fh;
  tdict_t *td;
  wchar_t *tag_delimiters = NULL;
  unsigned long line_nb;

  poptSetOtherOptionHelp(cl_context, "DICT");
  while ((cl_oh = poptGetNextOpt(cl_context)) > 0) {
    switch (cl_oh) {
    case CLOV_QUIET:
      verbosity = V_QUIET;
      break;
    case CLOV_VERBOSE:
      verbosity = V_VERBOSE;
      break;
    case CLOV_LOCACE:
      verbosity = V_LOCACE;
      break;
    case CLOV_STRICT:
      severity = V_WARNING;
      break;
    case CLOV_CI_ALL:
      search_options = search_options | PSEARCH_OPTION_CI_ALL;
      break;
    case CLOV_CI_FIRST:
      search_options = (search_options & (~PSEARCH_OPTION_CI_ALL)) | PSEARCH_OPTION_CI_FIRST;
      break;
    case CLOV_FORMAT_NO:
      if ((format_no = char2wcs(poptGetOptArg(cl_context))) == NULL)
	verbosity_message(V_QUIET | V_ERROR, "Illegal character sequence in --" CLOL_FORMAT_NO);
      break;
    case CLOV_FORMAT_YES:
       if ((format_yes = char2wcs(poptGetOptArg(cl_context))) == NULL)
	verbosity_message(V_QUIET | V_ERROR, "Illegal character sequence in --" CLOL_FORMAT_YES);
      break;
    case CLOV_GCI_ALL:
      search_options = search_options | PSEARCH_OPTION_GCI_ALL;
      break;
    case CLOV_GCI_FIRST:
      search_options = (search_options & (~PSEARCH_OPTION_GCI_ALL)) | PSEARCH_OPTION_GCI_FIRST;
      break;
    case CLOV_SUBSTRINGS:
      search_options = search_options & (~PSEARCH_OPTION_WHOLE);
      break;
    case CLOV_TAG_DELIMITERS:
      if ((tag_delimiters = char2wcs(poptGetOptArg(cl_context))) == NULL)
	verbosity_message(V_QUIET | V_ERROR, "Illegal character sequence in --" CLOL_TAG_DELIMITERS);
      break;
    case CLOV_VERSION:
      printf("%s %s\n", argv[0], VERSION);
      return 0;
    }
  }

  if (cl_oh != -1)
    verbosity_message(V_QUIET | V_ERROR, "%s %s", poptStrerror(cl_oh), poptBadOption(cl_context, 0));

  if (poptPeekArg(cl_context)) {
    dict_fn = poptGetArg(cl_context);
  }
  else {
    verbosity_message(V_QUIET | V_ERROR, "You must provide a dictionary file name");
  }

  poptFreeContext(cl_context);


  verbosity_message(V_NORMAL | V_INFO, "Opening dictionary file '%s'", dict_fn);
  dict_fh = open_file(dict_fn, 0);
  if (compiled_dict) {
    td = tdict_load(dict_fh);
  }
  else {
    wchar_t *line = calloc(dict_max_length + 1, sizeof(wchar_t));
    int nonls = -1;
    size_t tag_number = tabs_and_nls(tag_delimiters);
    wchar_t **tagset = calloc(tag_number, sizeof(wchar_t *));
    size_t l;

    verbosity_message(V_VERBOSE | V_INFO, "Creating dictionary (%s, tags: %u)", multi_dict ? "multi" : "single", tag_number);
    td = tdict_new(multi_dict, optimize_params[0], optimize_params[1], tag_number, optimize_params[2], optimize_params[3], optimize_params[4], optimize_params[5]);

    set_encoding(dict_encoding);

    if (headers) {
      verbosity_message(V_VERBOSE | V_INFO, "Reading header line");
      if (!fgetws(line, dict_max_length, dict_fh)) {
	if (errno == EILSEQ)
	  verbosity_message(V_QUIET | V_ERROR, "Illegal character sequence, line 1\n**** THIS ONE ---> %ls", line);
	else
	  verbosity_message(V_QUIET | V_ERROR, "Empty file?");
      }

      l = wcslen(line);
      nonls = get_nonls(line, l);
      line[l - nonls] = L'\0';
      if (!split_line(line, tag_delimiters, tagset))
	verbosity_message(V_QUIET | V_ERROR, "Not well formed header line");

      verbosity_message(V_VERBOSE | V_INFO, "Entry header: %ls", line);
      for (l = 0; l < tag_number; l++)
	verbosity_message(V_VERBOSE | V_INFO, "Tag header %u: %ls", l, tagset[l]);

      tdict_headers(td, line, tagset);
    }

    for (line_nb = tag_number ? 2 : 1; fgetws(line, dict_max_length, dict_fh) || (errno == EILSEQ); line_nb++) {
      if (errno == EILSEQ) {
	verbosity_message(V_NORMAL | V_WARNING, "Illegal character sequence, line %ld\n**** THIS ONE ---> %ls", line_nb, line);
	errno = 0;
	break;
      }
      l = wcslen(line);
      if (nonls < 0)
	nonls = get_nonls(line, l);
      line[l - nonls] = L'\0';
      if (!split_line(line, tag_delimiters, tagset)) {
	verbosity_message(V_NORMAL | V_WARNING, "Not well formed entry, line %ld\n**** THIS ONE ---> %ls", line_nb, line);
      }
      if (!tdict_add(td, line, tagset)) {
	verbosity_message(V_NORMAL | V_WARNING, "Duplicate entry, line %ld\n**** THIS ONE ---> %ls", line_nb, line);
      }
      errno = 0;
    }

    free(line);
    free(tagset);
  }
  fclose(dict_fh);

  if (save_as != NULL) {
    dict_fh = open_file(save_as, 1);
    verbosity_message(V_NORMAL | V_INFO, "Saving dictionary into file '%s'", save_as);
    tdict_save(td, dict_fh);
    fclose(dict_fh);
  }

  if (display_opt) {
    char *fmt = "--" CLOL_OPT_MNB " %d --" CLOL_OPT_MSB " %d --" CLOL_OPT_TNB " %d --" CLOL_OPT_TSB " %d --" CLOL_OPT_NENTRIES " %lu --" CLOL_OPT_TAGSIZE " %d\n";
    verbosity_message(V_NORMAL | V_INFO, "Computing optimal parameters");
    if (td->multi)
      printf(fmt,
	     BUNCH_CLAIMED(td->multimap->pat->nodes),
	     BUNCH_CLAIMED(td->multimap->pat->follow_strings),
	     BUNCH_CLAIMED(td->tag_collection->nodes),
	     BUNCH_CLAIMED(td->tag_collection->follow_strings),
	     td->value_number + td->tag_number + 1,
	     BUNCH_CLAIMED(td->multimap->values)
	     );
    else
      printf(fmt,
	     BUNCH_CLAIMED(td->map->nodes),
	     BUNCH_CLAIMED(td->map->follow_strings),
	     BUNCH_CLAIMED(td->tag_collection->nodes),
	     BUNCH_CLAIMED(td->tag_collection->follow_strings),
	     td->value_number + td->tag_number + 1,
	     BUNCH_CLAIMED(td->map->data)
	     );
  }

  if (!no_search) {
    FILE *in_fh = NULL;
    FILE *out_fh = NULL;
    bunch_t *display_bunch = bunch_new(sizeof(display_t), DEFAULT_DISPLAY_BLOCK);
    display_t *display_yes = NULL;
    display_t *display_no = NULL;
    wchar_t *line = NULL;
    wchar_t *saveline = NULL;
    size_t l;
    int nonls = -1;
    patricia_t *gci = NULL;
    psearch_t *search;

    tabs_and_nls(format_yes);
    display_yes = format_display(display_bunch, format_yes, 1, td->multi, td->tag_number);
    tabs_and_nls(format_no);
    display_no = format_display(display_bunch, format_no , 0, td->multi, td->tag_number);

    set_encoding(dict_encoding);

    if (search_options & PSEARCH_OPTION_GCI_ALL) {
      if (gci_fn == NULL) {
	verbosity_message(V_NORMAL | V_WARNING, "No GCI table provided, disabling GCI search");
	search_options = search_options & (~PSEARCH_OPTION_GCI_ALL);
      }
      else {
	FILE *gci_fh = open_file(gci_fn, 0);
	gci = gci_new(DEFAULT_GCI_NWC, DEFAULT_GCI_NALT);
	line = calloc(DEFAULT_GCI_MAX_LENGTH + 1, sizeof(wchar_t));
	for (line_nb = 1; fgetws(line, DEFAULT_GCI_MAX_LENGTH, gci_fh); line_nb++) {
	  l = wcslen(line);
	  if (nonls < 0)
	    nonls = get_nonls(line, l);
	  if (l - nonls) {
	    line[l - nonls] = L'\0';
	    if (!gci_add(gci, line[0], line + 1)) {
	      verbosity_message(V_NORMAL | V_WARNING, "Character already has alternatives, line %ld", line_nb);
	    }
	  }
	}
	free(line);
	line = NULL;
	nonls = -1;
      }
    }

    if (in_fn != NULL) {
      verbosity_message(V_VERBOSE | V_INFO, "Opening input file '%s'", in_fn);
      if (!(in_fh = fopen(in_fn, "r"))) {
	verbosity_message(V_QUIET | V_ERROR, "Could not open file '%s'", in_fn);
      }
    }
    else {
      in_fh = stdin;
    }

    if (out_fn != NULL) {
      verbosity_message(V_VERBOSE | V_INFO, "Opening file '%s' for writing", out_fn);
      if (!(out_fh = fopen(out_fn, "w"))) {
	verbosity_message(V_QUIET | V_ERROR, "Could not open '%s' for writing", out_fn);
      }
    }
    else {
      out_fh = stdout;
    }

    if (show_headers)
      display_headers(out_fh, td, display_yes);
    
    search = psearch_new((td->multi ? td->multimap->pat : td->map), search_options, gci, DEFAULT_OPT_DENSITY);
    line = calloc(search_max_length + 1, sizeof(wchar_t));
    
    verbosity_message(V_NORMAL | V_INFO, "Reading input");
    for (line_nb = 1; fgetws(line, search_max_length, in_fh); line_nb++) {
      l = wcslen(line);
      if (nonls < 0)
	nonls = get_nonls(line, l);
      if (l - nonls) {
	int alnum = 0;
	line[l - nonls] = L'\0';
	saveline = line;
	do {
	  if (!(words && alnum)) {
	    psearch_reinit(search);
	    psearch_new_query(search, saveline);
	    patricia_search(search);
	    if (search->final == NULL) {
	      display(out_fh, td, line, line_nb, NULL, display_no);
	    }
	    else {
	      for (search->current = search->final; search->current; search->current = search->next) {
		search->next = search->current->next;
		if (!(words && iswalpha(search->current->query[search->current->query_pos])))
		  display(out_fh, td, line, line_nb, search->current, display_yes);
	      }
	    }
	  }
	  alnum = iswalpha(*saveline);
	  saveline++;
	} while (substrings && *saveline);
      }
    }
    if (errno == EILSEQ) {
      verbosity_message(V_NORMAL | V_WARNING, "Illegal character sequence, line %ld", line_nb);
    }

    free(line);
    free(format_yes);
    free(format_no);
    psearch_free(search);
    bunch_free(display_bunch);
  }

  tdict_free(td);
  
  return 0;
}
