/*
  Copyright (C) 1988-1994,1996-1999,2003,2004 Free Software Foundation, Inc.
  This file is part of the GNU C Library.  Its master source is NOT part of
  the C library, however.  The master source lives in /gd/gnu/lib.

  The GNU C Library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  The GNU C Library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with the GNU C Library; if not, write to the Free
  Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
  02111-1307 USA.
*/

#include "trish_dict.h"
#include "string_collection.h"

#include "mouchard.h"

tdict_t *tdict_new(int multi, unsigned int main_node_block, unsigned int main_string_block, int n_tags, unsigned int tag_node_block, unsigned int tag_string_block, unsigned int entry_number, unsigned int tag_size) {
  tdict_t *td = calloc(1, sizeof(tdict_t));
  bunch_t *data = bunch_new(sizeof(wchar_t *), entry_number * n_tags);

  td->multi = multi;
  td->entry_number = 0;
  td->value_number = 0;
  td->tag_number = n_tags;
  if (multi)
    td->multimap = pmap_multi_new(main_node_block, main_string_block, data);
  else
    td->map = patricia_new(main_node_block, main_string_block, data);

  td->tag_collection = string_collection_new(tag_node_block, tag_string_block, tag_size);

  return td;
}

void tdict_free(tdict_t *td) {
  if (td->multi) {
    bunch_free(td->multimap->values);
    pmap_multi_free(td->multimap);
  }
  else {
    bunch_free(td->map->data);
    patricia_free(td->map);
  }

  bunch_free(td->tag_collection->data);

  if (td->tag_collection->nodes)
    patricia_free(td->tag_collection);
  else
    free(td->tag_collection);

  free(td);
}

void tdict_save(tdict_t *td, FILE *f) {
  fwrite(td, sizeof(tdict_t), 1, f);
  if (td->tag_headers != NULL)
    fwrite(td->tag_headers, sizeof(wchar_t *), td->tag_number, f);
  if (td->multi)
    pmap_multi_save(td->multimap, BUNCH_ATTRIB_PRUNED_BLOCKS, f);
  else
    patricia_save(td->map, BUNCH_ATTRIB_PRUNED_BLOCKS, f);
  bunch_save(td->tag_collection->data, BUNCH_ATTRIB_PRUNED_BLOCKS, f);
}

tdict_t *tdict_load(FILE *f) {
  tdict_t *td = calloc(1, sizeof(tdict_t));
  int i, j;

  fread(td, sizeof(tdict_t), 1, f);
  if (td->tag_headers != NULL) {
    td->tag_headers = calloc(td->tag_number, sizeof(wchar_t *));
    fread(td->tag_headers, sizeof(wchar_t *), td->tag_number, f);
  }
  td->tag_collection = calloc(1, sizeof(patricia_t));
  if (td->multi) {
    td->multimap = pmap_multi_load(f);
    td->tag_collection->data = bunch_load(f);
    for (i = 0; i < td->multimap->values->block_number; i++)
      for (j = 0; j < td->multimap->values->block_size - td->multimap->values->blocks[i]->unclaimed; j++) {
	wchar_t **p = (wchar_t **)td->multimap->values->blocks[i]->data + j;
	BUNCH_CORRECT_PTR(td->tag_collection->data, *p);
      }
  }
  else {
    td->map = patricia_load(f);
    td->tag_collection->data = bunch_load(f);
    for (i = 0; i < td->map->data->block_number; i++)
      for (j = 0; j < td->map->data->block_size - td->map->data->blocks[i]->unclaimed; j++) {
	wchar_t **p = (wchar_t **)td->map->data->blocks[i]->data + j;
	BUNCH_CORRECT_PTR(td->tag_collection->data, *p);
      }
  }

  if (td->entry_header != NULL)
    BUNCH_CORRECT_PTR(td->tag_collection->data, td->entry_header);
  if (td->tag_headers != NULL)
    for (i = 0; i < td->tag_number; i++)
      BUNCH_CORRECT_PTR(td->tag_collection->data, td->tag_headers[i]);

  return td;
}

int tdict_add(tdict_t *td, wchar_t *entry, wchar_t **tagset) {
  wchar_t **claimed_tagset;
  int i;

  claimed_tagset = bunch_claim((td->multi ? td->multimap->values : td->map->data), td->tag_number);
  for (i = 0; i < td->tag_number; i++) {
    claimed_tagset[i] = string_collection_ptr(td->tag_collection, tagset[i]);
  }

  if (td->multi)
    pmap_multi_add(td->multimap, entry, claimed_tagset);
  else
    if (pmap_add(td->map, entry, claimed_tagset, 0))
      return 0;

  td->value_number++;

  return 1;
}

void tdict_headers(tdict_t *td, wchar_t *entry, wchar_t **tagset) {
  int i;

  td->tag_headers = bunch_claim((td->multi ? td->multimap->values : td->map->data), td->tag_number);
  for (i = 0; i < td->tag_number; i++) {
    td->tag_headers[i] = string_collection_ptr(td->tag_collection, tagset[i]);
  }
  td->entry_header = string_collection_ptr(td->tag_collection, entry);
}
