/*
  trish_dict.h - dictionary structure common to trish-* utils, part of libtrish

  Copyright (C) 1988-1994,1996-1999,2003,2004 Free Software Foundation, Inc.
  This file is part of the GNU C Library.  Its master source is NOT part of
  the C library, however.  The master source lives in /gd/gnu/lib.

  The GNU C Library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  The GNU C Library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with the GNU C Library; if not, write to the Free
  Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
  02111-1307 USA.
*/

#ifndef HAVE_TRISH_DICT_H
#define HAVE_TRISH_DICT_H

#include "patricia.h"
#include "patricia_map.h"


/*
  tdict_t - dictionary structure
*/

typedef struct {
  /* multi - if non-zero, each entry may have several tagsets */
  int multi;

  /* entry_number - number of different entries XXX */
  unsigned long entry_number;

  /* value_number - number of entries, may be higher than entry_number in multi dicts */
  unsigned long value_number;

  /* map - PATRICIA tree holding entries, NULL if multi is non-zero */
  patricia_t *map;

  /* multimap - PATRICIA tree holding entries, NULL if multi is zero */
  pmap_multi_t *multimap;

  /* tag_number - number of tags for each entry */
  int tag_number;

  /* tag_headers - name of each tag */
  wchar_t **tag_headers;

  /* entry_header - name of the entry string */
  wchar_t *entry_header;

  /* tag_collection - string collection holding all tags */
  patricia_t *tag_collection;
} tdict_t;


/*
  tdict_new - creates a new dictionary
  multi - non-zero if it allows several tagsets for a single entry
  main_node_block - block size for the entries node bunch
  main_string_block - block size for the entries follow string bunch
  n_tags - number of tags for each entry
  tag_node_block - block size for the tags node bunch
  tag_string_block - block size for the tags follow string bunch
  entry_number - estimated number of entries
  tag_size - estimated size of the tag space in characters
  return value - a frshly allocated new dictionary
*/

tdict_t *tdict_new(int multi, unsigned int main_node_block, unsigned int main_string_block, int n_tags, unsigned int tag_node_block, unsigned int tag_string_block, unsigned int entry_number, unsigned int tag_size);


/*
  tdict_free - destroys a dictionary
  td - dictionary to destroy

  Note: this function will free EVERYTHING
*/

void tdict_free(tdict_t *td);


/*
  tdict_save - writes a dictionary into a file
  td - dictionary to write
  f - file handle to write into, must be opened with fopen with mode "w" or "a"

  Note: every bunch will be pruned
*/

void tdict_save(tdict_t *td, FILE *f);


/*
  tdict_load - reads a dictionary from a file
  f - file handle to read from, must be opened with fopen with mode "r"
  return value - the dictionary

  Note: you cannot add an entry to a loaded dictionary
*/

tdict_t *tdict_load(FILE *f);


/*
  tdict_add - adds an entry to a dictionary
  td - dictionary
  entry - the entry string
  tagset - an array of tags, it must be the same size than td->tag_number
  return value - a status value, non-zero on success, zero if td allows only one tagset and the entry already exists
*/

int tdict_add(tdict_t *td, wchar_t *entry, wchar_t **tagset);


/*
  tdict_headers - sets the tag header names for a dictionary
  td - dictionary
  entry - the entry header
  tagset - an array of headers, it must be the same size than td->tag_number
*/

void tdict_headers(tdict_t *td, wchar_t *entry, wchar_t **tagset);

#endif
