#include "trkov_defaults.h"
#include "trkov_cl_config.h"
#include "verbosity.h"
#include "trkov_dict.h"
#include "wchar_utils.h"
#include "patricia_search.h"

#include <popt.h>
#include <locale.h>
#include <wchar.h>
#include <errno.h>
#include <time.h>
#include <string.h>

#include "mouchard.h"

FILE *open_file(const char *fn, int write) {
  FILE *fh = fopen(fn, write ? "w" : "r");

  if (fh == NULL)
    verbosity_message(VERBOSITY_ERROR, "Could not open file '%s'", fn);

  return fh;
}

void set_encoding(char *encoding) {
  if (setlocale(LC_CTYPE, encoding) == NULL) {
    verbosity_message(VERBOSITY_ERROR, "Unhandled locale '%s'", encoding);
  }
  verbosity_message(VERBOSITY_INFO, "Locale set to '%s'", encoding);
}

typedef struct _trk_pattern_t trk_pattern_t;
struct _trk_pattern_t {
  int command;
  union {
    wchar_t wc;
    int k;
    trk_pattern_t *back;
  } arg;
  trk_pattern_t *out;
  trk_pattern_t *next;
};

#define TRK_COMMAND_NOP         0
#define TRK_COMMAND_FIXED       1
#define TRK_COMMAND_GLOBAL      2
#define TRK_COMMAND_POSITIONAL  3
#define TRK_COMMAND_BLOCK_END   4

#define TRK_COMMAND_WCHAR_GLOBAL       L'-'
#define TRK_COMMAND_WCHAR_POSITIONAL   L'~'
#define TRK_COMMAND_WCHAR_BLOCK_START  L'('
#define TRK_COMMAND_WCHAR_BLOCK_END    L')'

trk_pattern_t *trk_pattern_compile(const wchar_t *w, bunch_t *b, trk_dict_t *trkd) {
  int offset;
  int esc = 0;
  int err = 0;
  int back = 0;
  trk_pattern_t *pattern = bunch_claim(b, 1);
  trk_pattern_t *current = pattern;
  trk_pattern_t *i;
  trk_pattern_t *block = NULL;

  pattern->command = TRK_COMMAND_NOP;

#define NEW_PATTERN { i = bunch_claim(b, 1); current->next = i; current = i; i = NULL; }
#define SOERR {\
  err++;\
  verbosity_message(VERBOSITY_INFO | V_NONL, "Pattern character %d: ", offset);\
  verbosity_message(VERBOSITY_INFO | V_NOHDR,
#define EOERR ); }

  for (offset = 0; w[offset]; offset++) {
    if (esc) {
      NEW_PATTERN;
      current->command = TRK_COMMAND_FIXED;
      current->arg.wc = w[offset];
      esc = 0;
    }
    else {
      switch (w[offset]) {
      case TRK_COMMAND_WCHAR_GLOBAL:
	NEW_PATTERN;
	current->command = TRK_COMMAND_GLOBAL;
	offset++;
	current->arg.k = w[offset] - L'0';
	if (current->arg.k < 0) {
	  SOERR "Request for negative order %d", current->arg.k EOERR;
	}
	if (trkd->kmax && (current->arg.k > trkd->kmax)) {
	  SOERR "Request for order %d whereas the dictionary has an inferior kmax of %d", current->arg.k, trkd->kmax EOERR;
	}
	break;
      case TRK_COMMAND_WCHAR_POSITIONAL:
	NEW_PATTERN;
	current->command = TRK_COMMAND_POSITIONAL;
	offset++;
	current->arg.k = w[offset] - L'0';
	if (trkd->positional == NULL) {
	  SOERR "Request for positional generation whereas the dictionary has no positional matrix" EOERR;
	}
	if (current->arg.k < 0) {
	  SOERR "Request for negative order %d", current->arg.k EOERR;
	}
	if (trkd->kmax && (current->arg.k > trkd->kmax)) {
	  SOERR "Request for order %d whereas the dictionary has an inferior kmax of %d", current->arg.k, trkd->kmax EOERR;
	}
	break;
      case TRK_COMMAND_WCHAR_BLOCK_START:
	if (back) {
	  SOERR "Extra opening block" EOERR;
	}
	else {
	  block = current;
	  back = 1;
	}
	break;
      case TRK_COMMAND_WCHAR_BLOCK_END:
	NEW_PATTERN;
	current->command = TRK_COMMAND_BLOCK_END;
	current->arg.back = block;
	if (!back) {
	  SOERR "Opening block is missing" EOERR;
	  break;
	}
	if (block == NULL) {
	  SOERR "Extra closing block" EOERR;
	  break;
	}
	else
	  for (i = block; i != current; i = i->next)
	    i->out = current;
	block = NULL;
	break;
      case ESCAPE_WCHAR:
	esc = 1;
	break;
      default:
	NEW_PATTERN;
	current->command = TRK_COMMAND_FIXED;
	current->arg.wc = w[offset];
	break;
      }
    }
  }

  if (esc) {
    SOERR "Escape at the end of pattern" EOERR;
  }

  if (err) {
    verbosity_message(V_QUIET | V_ERROR, "There are %d error(s) in the pattern", err);
  }

#undef NEW_PATTERN
#undef SOERR
#undef EOERR
  return pattern;
}



typedef struct {
  wchar_t *w;
  size_t length;
  float p;
} trk_word_t;

pnode_t *trk_lookup(patricia_t *pat, const wchar_t *word, int start, int *follow_pos) {
  pnode_t *node = pat->root;

  *follow_pos = 0;

  if (start)
    node = pnode_lookup(node, PNODE_TYPE_START, L'\0');

  while (*word && (node != NULL)) {
    if ((node->follow_string == NULL) || !node->follow_string[*follow_pos]) {
      node = pnode_lookup(node, PNODE_TYPE_CHAR, *word);
      *follow_pos = 0;
    }
    else {
      if (*word == node->follow_string[*follow_pos]) {
	*follow_pos += 1;
      }
      else {
	*follow_pos = 0;
	return NULL;
      }
    }
    word++;
  }

  return node;
}

int pnode_children(pnode_t *node) {
  if (node == NULL)
    return 0;
  return 1 + pnode_children(node->left) + pnode_children(node->right);
}

pnode_t *trk_pick(pnode_t *node, uintptr_t *luck, int equiprobable) {
  if (node == NULL) return NULL;

  if (node->left != NULL) {
      pnode_t *node2 = trk_pick(node->left, luck, equiprobable);
      if (node2 != NULL)
	return node2;
  }
  if (node->pnode_type != PNODE_TYPE_START) {
    if (equiprobable) {
      if (*luck < 1)
	return node;
      *luck -= 1;
    }
    else {
      if (*luck < (uintptr_t)node->data)
	return node;
      *luck -= (uintptr_t)node->data;
    }
  }
  if (node->right != NULL) {
      pnode_t *node2 = trk_pick(node->right, luck, equiprobable);
      if (node2 != NULL)
	return node2;
  }

  return NULL;
}

#define TRK_PUT_CHAR(gen, wc) { (gen)->w[(gen)->length] = (wc);\
      (gen)->length++;\
      (gen)->w[(gen)->length] = L'\0';\
}

void trk_generate(trk_word_t *gen, trk_dict_t *dict, trk_pattern_t *pattern, int max_length, int equiprobable) {
  gen->w[0] = L'\0';
  gen->length = 0;
  gen->p = 1.0;

  while ((pattern != NULL) && (gen->length < max_length)) {
    int off;
    pnode_t *n = NULL;
    int follow_pos = 0;
    uintptr_t luck;
    
    switch (pattern->command) {
    case TRK_COMMAND_NOP:
      break;
    case TRK_COMMAND_FIXED:
      TRK_PUT_CHAR(gen, pattern->arg.wc);
      break;
    case TRK_COMMAND_GLOBAL:
      off = gen->length - ((gen->length > pattern->arg.k) ? pattern->arg.k : gen->length);

      n = trk_lookup(dict->global, gen->w + off, !off /* XXX option*/, &follow_pos);
      while ((n == NULL) && (off < gen->length)) {
	n = trk_lookup(dict->global, gen->w + off, 0, &follow_pos);
	off++;
      }

      if ((n->follow_string == NULL) || !n->follow_string[follow_pos]) {
	int pnc = pnode_children(n->next);
	luck = rand() % (equiprobable ? pnc : PNODE_COUNT(n));
	n = trk_pick(n->next, &luck, equiprobable);

	gen->p *= (float)PNODE_COUNT(n) / (float)PNODE_COUNT(n->parent);

	if ((n == NULL) || (n->pnode_type == PNODE_TYPE_END)) {
	  pattern = pattern->out;
	}
	else {
	  TRK_PUT_CHAR(gen, n->wc);
	}
      }
      else {
	TRK_PUT_CHAR(gen, n->follow_string[follow_pos]);
      }
      break;
    case TRK_COMMAND_POSITIONAL:
      off = gen->length - ((gen->length > pattern->arg.k) ? pattern->arg.k : gen->length);

      n = trk_lookup(dict->positional[off], gen->w + off, 0, &follow_pos);
      while ((n == NULL) && (off < gen->length)) {
	n = trk_lookup(dict->positional[off], gen->w + off, 0, &follow_pos);
	off++;
      }

      if ((n->follow_string == NULL) || !n->follow_string[follow_pos]) {
	luck = rand() % (equiprobable ? pnode_children(n->next) : PNODE_COUNT(n));
	n = trk_pick(n->next, &luck, equiprobable);

	gen->p *= (float)PNODE_COUNT(n) / (float)PNODE_COUNT(n->parent);

	if ((n == NULL) || (n->pnode_type == PNODE_TYPE_END)) {
	  pattern = pattern->out;
	}
	else {
	  gen->w[gen->length] = n->wc;
	  gen->length++;
	  gen->w[gen->length] = L'\0';
	}
      }
      else {
	gen->w[gen->length] = n->follow_string[follow_pos];
	gen->length++;
	gen->w[gen->length] = L'\0';
      }
      break;
    case TRK_COMMAND_BLOCK_END:
      pattern = pattern->arg.back;
      break;
    }

    if (pattern != NULL)
      pattern = pattern->next;
  }
}

typedef struct _trk_fit_t trk_fit_t;
struct _trk_fit_t {
  float p;
  trk_pattern_t *pattern;
  int pos;
  trk_fit_t *next;
  trk_fit_t *prev;
};

float trk_fit(trk_dict_t *dict, trk_pattern_t *pattern, wchar_t *w) {
  bunch_t *fitb = bunch_new(sizeof(trk_fit_t), 100 /* XXX */);
  trk_fit_t *first = bunch_claim(fitb, 1);
  trk_fit_t *current = first;
  float p = 0.0;
  first->p = 1.0;
  first->pattern = pattern;

#define DISCARD_FIT {\
  if (current == first) {\
    first = current->next;\
  }\
  if (current->prev != NULL) {\
    current->prev->next = current->next;\
  }\
  if (current->next != NULL) {\
    current->next->prev = current->prev;\
  }\
}

  while (1) {
    if (!w[current->pos]) {
      p = current->p;
      break;
    }
    if (current->pattern == NULL) {
      DISCARD_FIT;
    }
    else {
      switch (current->pattern->command) {
	wchar_t wc;
	int off;
	pnode_t *n;
	int follow_pos;
      case TRK_COMMAND_NOP:
	current->pattern = current->pattern->next;
	break;
      case TRK_COMMAND_FIXED:
	if (w[current->pos] == current->pattern->arg.wc) {
	  current->pattern = current->pattern->next;
	  current->pos += 1;
	}
	else {
	  DISCARD_FIT;
	}
	break;
      case TRK_COMMAND_GLOBAL:
	wc = w[current->pos];
	w[current->pos] = L'\0';
	off = current->pos - ((current->pos > current->pattern->arg.k) ? current->pattern->arg.k : current->pos);
	n = trk_lookup(dict->global, w + off, !off /* XXX option*/, &follow_pos);
	w[current->pos] = wc;
	if (n == NULL) {
	  DISCARD_FIT;
	}
	else {
	  if ((n->follow_string == NULL) || !n->follow_string[follow_pos]) {
	    pnode_t *m = pnode_lookup(n, PNODE_TYPE_CHAR, wc);
	    if (m == NULL) {
	      DISCARD_FIT;
	    }
	    else {
	      current->p *= (float)PNODE_COUNT(m) / (float)PNODE_COUNT(n);
	      
	      if (current->pattern->out != NULL) {
		m = pnode_lookup(n, PNODE_TYPE_END, L'\0');
		if (m != NULL) {
		  trk_fit_t *new = bunch_claim(fitb, 1);
		  new->p = current->p * (float)PNODE_COUNT(m) / (float)PNODE_COUNT(n);
		  new->pattern = current->pattern->out->next;
		  new->pos = current->pos;
		  new->prev = current;
		  new->next = current->next;
		  current->next = new;
		  if (new->next != NULL)
		    new->next->prev = new;
		}
	      }

	      current->pattern = current->pattern->next;
	      current->pos += 1;
	    }
	  }
	  else {
	    if (w[current->pos] == n->follow_string[follow_pos]) {
	      current->pattern = current->pattern->next;
	      current->pos += 1;
	    }
	    else {
	      DISCARD_FIT;
	    }
	  }
	}
	break;
      case TRK_COMMAND_POSITIONAL:
	break;
      case TRK_COMMAND_BLOCK_END:
	current->pattern = current->pattern->arg.back;
	break;
      }
    }
    if (current->next == NULL) {
      if (first == NULL) {
	p = 0.0;
	break;
      }
      current = first;
    }
    else {
      current = current->next;
    }
  }

  bunch_free(fitb);
  return p;
}




int main(int argc, const char **argv) {
  static int compiled = 0;
  static int max_length = DEFAULT_MAX_LENGTH;
  static int positional = 0;
  static int kmax = 0;
  static char *dict_encoding = DEFAULT_ENCODING;
  static char *out_fn = NULL;
  static char *gen_encoding = DEFAULT_ENCODING;
  static int show_proba = 0;
  static char *seed_src = DEFAULT_SEED_SRC;
  static int number = DEFAULT_TRKOV_NUMBER;
  static int hide_seed = 0;
  static char *saveas = NULL;
  static int filter = 0;
  static int fit = 0;
  static int equiprobable = 0;

  static struct poptOption cl_verbosity_options[] = CL_VERBOSITY_OPTIONS;

  static struct poptOption cl_dict_options[] = {
    { CLOL_COMPILED      , CLOS_COMPILED      , POPT_ARG_NONE   , &compiled      , 0 , CLOH_COMPILED      , NULL },
    { CLOL_SAVE_AS       , CLOS_SAVE_AS       , POPT_ARG_STRING , &saveas        , 0 , CLOH_SAVE_AS       , "FILE" },
    { CLOL_MAX_LENGTH    , CLOS_MAX_LENGTH    , POPT_ARG_INT    , &max_length    , 0 , CLOH_MAX_LENGTH    , "LENGTH" },
    { CLOL_POSITIONAL    , CLOS_POSITIONAL    , POPT_ARG_NONE   , &positional    , 0 , CLOH_POSITIONAL    , NULL },
    { CLOL_DICT_ENCODING , CLOS_DICT_ENCODING , POPT_ARG_STRING , &dict_encoding , 0 , CLOH_DICT_ENCODING , "LOCALE" },
    { CLOL_KMAX          , CLOS_KMAX          , POPT_ARG_INT    , &kmax          , 0 , CLOH_KMAX          , "KMAX" },
    POPT_TABLEEND
  };

  static struct poptOption cl_gen_options[] = {
    { CLOL_OUTPUT       , CLOS_OUTPUT       , POPT_ARG_STRING , &out_fn       , 0         , CLOH_OUTPUT       , "FILE" },
    { CLOL_GEN_ENCODING , CLOS_GEN_ENCODING , POPT_ARG_STRING , &gen_encoding , 0         , CLOH_GEN_ENCODING , "LOCALE" },
    { CLOL_SHOW_PROBA   , CLOS_SHOW_PROBA   , POPT_ARG_NONE   , &show_proba   , 0         , CLOH_SHOW_PROBA   , NULL },
    { CLOL_EQUIPROBABLE , CLOS_EQUIPROBABLE , POPT_ARG_NONE   , &equiprobable , 0         , CLOH_EQUIPROBABLE , NULL },
    { CLOL_SEED         , CLOS_SEED         , POPT_ARG_STRING , &seed_src     , 0         , CLOH_SEED         , "INT" },
    { CLOL_NUMBER       , CLOS_NUMBER       , POPT_ARG_INT    , &number       , 0         , CLOH_NUMBER       , "INT" },
    { CLOL_HIDE_SEED    , CLOS_HIDE_SEED    , POPT_ARG_NONE   , &hide_seed    , 0         , CLOH_HIDE_SEED    , NULL },
    { CLOL_FILTER_DICT  , CLOS_FILTER_DICT  , POPT_ARG_NONE   , &filter       , 0         , CLOH_FILTER_DICT  , NULL },
    POPT_TABLEEND
  };

  static struct poptOption cl_options[] = {
    POPT_AUTOHELP
    { CLOL_FIT           , CLOS_FIT           , POPT_ARG_NONE   , &fit           , 0                  , CLOH_FIT           , NULL },
    { CLOL_VERSION       , CLOS_VERSION       , POPT_ARG_NONE   , NULL           , CLOV_VERSION       , CLOH_VERSION       , NULL },
    { NULL , '\0' , POPT_ARG_INCLUDE_TABLE , cl_verbosity_options , 0 , "Verbosity options:" , NULL },
    { NULL , '\0' , POPT_ARG_INCLUDE_TABLE , cl_dict_options , 0 , "Dictionary options:" , NULL },
    { NULL , '\0' , POPT_ARG_INCLUDE_TABLE , cl_gen_options , 0 , "Generation options:" , NULL },
    POPT_TABLEEND
  };

  poptContext cl_context = poptGetContext(NULL, argc, argv, cl_options, 0);
  int cl_oh;

  wchar_t *pattern_w = NULL;
  const char *dict_fn = NULL;
  FILE *dict_fh;
  FILE *out_fh = NULL;
  trk_dict_t *dict;
  bunch_t *patb = bunch_new(sizeof(trk_pattern_t), DEFAULT_PATTERN_LENGTH);
  trk_pattern_t *pattern = NULL;
  psearch_t *search = NULL;
  int seed;

  set_encoding(getenv("LC_CTYPE"));
  poptSetOtherOptionHelp(cl_context, "PATTERN DICT");
  while ((cl_oh = poptGetNextOpt(cl_context)) > 0) {
    switch (cl_oh) {
    case CLOV_QUIET:
      verbosity = V_QUIET;
      break;
    case CLOV_VERBOSE:
      verbosity = V_VERBOSE;
      break;
    case CLOV_LOCACE:
      verbosity = V_LOCACE;
      break;
    case CLOV_STRICT:
      severity = V_WARNING;
      break;
    case CLOV_VERSION:
      printf("%s %s\n", argv[0], VERSION);
      return 0;
    }
  }

  if (cl_oh != -1)
    verbosity_message(V_QUIET | V_ERROR, "%s %s", poptStrerror(cl_oh), poptBadOption(cl_context, 0));

  if (poptPeekArg(cl_context)) {
    const char *pattern_s = poptGetArg(cl_context);
    if ((pattern_w = char2wcs(pattern_s)) == NULL) {
      verbosity_message(V_QUIET | V_ERROR, "Illegal character sequence in pattern");
    }
  }
  else {
    verbosity_message(VERBOSITY_ERROR, "You must provide a pattern");
  }

  if (poptPeekArg(cl_context)) {
    dict_fn = poptGetArg(cl_context);
  }
  else {
    verbosity_message(VERBOSITY_ERROR, "You must provide a dictionary file name");
  }

  poptFreeContext(cl_context);

  set_encoding(dict_encoding);
  if ((dict_fh = open_file(dict_fn, 0)) == NULL) {
    verbosity_message(V_QUIET | V_ERROR, "Could not open file '%s'", dict_fn);
  }
  verbosity_message(V_NORMAL | V_INFO, "Opening file '%s'", dict_fn);
  if (compiled) {
    dict = trk_dict_load(dict_fh);
  }
  else {
    wchar_t *line = calloc(max_length + 1, sizeof(wchar_t));
    int nonls = -1;
    size_t l;
    unsigned long line_nb; 

    dict = trk_dict_new(1500 /* XXX */, kmax, positional);
  
    for (line_nb = 1; fgetws(line, max_length, dict_fh) || (errno == EILSEQ); line_nb++) {
      if (errno == EILSEQ) {
        verbosity_message(V_NORMAL | V_WARNING, "Illegal character sequence, line %ld\n**** THIS ONE ---> %ls", line_nb, line);
        errno = 0;
        break;
      }
      l = wcslen(line);
      if (nonls < 0)
        nonls = get_nonls(line, l);
      line[l - nonls] = L'\0';
      trk_dict_add(dict, line);
    }

    free(line);
  }
  fclose(dict_fh);

  if (saveas != NULL) {
    FILE *f = open_file(saveas, 1);
    verbosity_message(V_NORMAL | V_INFO, "Saving dictionary into '%s'", saveas);
    trk_dict_save(dict, f);
    fclose(f);
  }

  pattern = trk_pattern_compile(pattern_w, patb, dict);
  free(pattern_w);

  if (fit) {
    wchar_t *line = calloc(max_length + 1, sizeof(wchar_t));
    int nonls = -1;
    size_t l;
    unsigned long line_nb; 
    float p;

    verbosity_message(V_NORMAL | V_INFO, "Reading words from stdin");
    for (line_nb = 1; fgetws(line, max_length, stdin) || (errno == EILSEQ); line_nb++) {
      if (errno == EILSEQ) {
        verbosity_message(V_NORMAL | V_WARNING, "Illegal character sequence, line %ld\n**** THIS ONE ---> %ls", line_nb, line);
        errno = 0;
        break;
      }
      l = wcslen(line);
      if (nonls < 0)
        nonls = get_nonls(line, l);
      line[l - nonls] = L'\0';
      p = trk_fit(dict, pattern, line);
      printf("%ls\t%.12f\n", line, p);
    }
    free(line);
  }
  else {
    trk_word_t gen;
    int i;
 
    if (!strcspn(seed_src, "0123456789")) {
      seed = atoi(seed_src);
    }
    else if (!strcmp(seed_src, "time")) {
      seed = time(NULL);
    }
    else {
      FILE *f = open_file(seed_src, 0);
      fread(&seed, sizeof(unsigned int), 1, f);
      fclose(f);
    }
    srand(seed);
    if (!hide_seed) {
      printf("seed = %u\n", seed);
    }
    
    set_encoding(gen_encoding);
    if (out_fn == NULL) {
      out_fh = stdout;
    }
    else 
      out_fh = open_file(out_fn, 1);
    
    gen.w = calloc(MAX_GENERATE_LENGTH + 1, sizeof(wchar_t));
    if (filter) {
      search = psearch_new(dict->global, PSEARCH_OPTION_FULL, NULL, 1);
    }
    for (i = 0; i < number; i++) {
      trk_generate(&gen, dict, pattern, MAX_GENERATE_LENGTH, equiprobable);
      if (filter) {
	psearch_reinit(search);
	psearch_new_query(search, gen.w);
	patricia_search(search);
	if (search->final) {
	  verbosity_message(V_VERBOSE | V_INFO, "Filtered out '%ls'", gen.w);
	  i--;
	  continue;
	}
      }
      if (show_proba) {
	fprintf(out_fh, "%ls\t%.12f\n", gen.w, gen.p);
      }
      else {
	fprintf(out_fh, "%ls\n", gen.w);
      }
    }
    if (filter) {
      psearch_free(search);
    }
    free(gen.w);
  }

  trk_dict_free(dict);
  bunch_free(patb);

  return 0;
}
