#include "trkov_dict.h"
#include "trkov_defaults.h"

#include <wchar.h>

#include "mouchard.h"
#include "wchar_utils.h"

unsigned int estimate_node_block(unsigned int nentries, int max_length) {
  return (nentries * (max_length + 1)) / 10;
}

unsigned int estimate_string_block(unsigned int nentries, int max_length) {
  return (nentries * (max_length + 1)) / 10;
}

trk_dict_t *trk_dict_new(unsigned int nentries, int kmax, int pos) {
  trk_dict_t *tk = calloc(1, sizeof(trk_dict_t));
  unsigned int nb = estimate_node_block(nentries, kmax || DEFAULT_MAX_LENGTH);
  unsigned int sb = estimate_string_block(nentries, kmax || DEFAULT_MAX_LENGTH);

  tk->kmax = kmax;
  tk->global = patricia_new(nb, sb, NULL);
  tk->max_pos = 0;

  if (pos) {
    tk->positional = calloc(1, sizeof(patricia_t *));
    tk->positional[0] = patricia_new(nb, sb, NULL);
  }
  else {
    tk->positional = NULL;
  }

  return tk;
}

void trk_dict_free(trk_dict_t *tk) {
  patricia_free(tk->global);
  if (tk->positional) {
    int i;
    for (i = 0; i < tk->max_pos; i++) {
      patricia_free(tk->positional[i]);
   }
    free(tk->positional);
  }
  free(tk);
}

void trk_dict_save(trk_dict_t *tk, FILE *f) {
  int i;
  fwrite(tk, sizeof(trk_dict_t), 1, f);
  patricia_save(tk->global, BUNCH_ATTRIB_PRUNED_BLOCKS, f);
  for (i = 0; i < tk->max_pos; i++)
    patricia_save(tk->positional[i], BUNCH_ATTRIB_PRUNED_BLOCKS, f);
}

trk_dict_t *trk_dict_load(FILE *f) {
  trk_dict_t *tk = calloc(1, sizeof(trk_dict_t));

  fread(tk, sizeof(trk_dict_t), 1, f);
  tk->global = patricia_load(f);
  if (tk->positional != NULL) {
    int i;
    tk->positional = calloc(tk->max_pos, sizeof(patricia_t *));
    for (i = 0; i < tk->max_pos; i++) {
      tk->positional[i] = patricia_load(f);
    }
  }

  return tk;
}

void trk_count_aux(patricia_t *pat) {
  unsigned int i;

  for (i = 0; i < pat->nodes->block_number; i++) {
    unsigned int j;
    for (j = 0; j < pat->nodes->block_size - pat->nodes->blocks[i]->unclaimed; j++) {
      pnode_t *n = ((pnode_t *)pat->nodes->blocks[i]->data) + j;
      if (n->next == NULL) {
	while (n != NULL) {
	  n->data = (void *)((uintptr_t)n->data + 1);
	  n = n->parent;
	}
      }
    }
  }
}

#define PNODE_INCR(node) (node)->data = (void *)(PNODE_COUNT(node) + 1)

void trk_count(const patricia_add_context_t *ctx) {
  switch (ctx->event) {
  case PATRICIA_ADD_EVENT_FOLLOW_STRING:
    if ((!ctx->word[ctx->word_pos]) && !(ctx->options & PATRICIA_ADD_OPTION_END)) {
      PNODE_INCR(ctx->current);
    }
    break;
  case PATRICIA_ADD_EVENT_NEXT_NODE:
  case PATRICIA_ADD_EVENT_NEW_NODE:
    if (ctx->current->pnode_type != PNODE_TYPE_START) {
      PNODE_INCR(ctx->parent);
    }
    if ((ctx->current->pnode_type == PNODE_TYPE_END) || ((!ctx->word[ctx->word_pos]) && !(ctx->options & PATRICIA_ADD_OPTION_END))){
      PNODE_INCR(ctx->current);
    }
    break;
  case PATRICIA_ADD_EVENT_SPLIT_FOLLOW_STRING:
    ctx->current->data = (void *)PNODE_COUNT(ctx->parent);
    break;
  }
}

void trk_dict_add(trk_dict_t *tk, wchar_t *word) {
  patricia_add_context_t ctx;
  int i;
  size_t l = wcslen(word);

  if ((tk->positional != NULL) && (l > tk->max_pos)) {
    tk->positional = realloc(tk->positional, (l + 1) * sizeof(patricia_t *));
    for (i = tk->max_pos + 1; i <= l; i++) {
      tk->positional[i] = patricia_new(tk->positional[i - 1]->nodes->block_size, tk->positional[i - 1]->follow_strings->block_size, NULL);
    }
    tk->max_pos = l;
  }

  patricia_add_context_init(&ctx, tk->global, word, PATRICIA_ADD_OPTION_START | PATRICIA_ADD_OPTION_END, NULL);
  patricia_add(&ctx, trk_count);

  for (i = 0; i < l; i++) {
    wchar_t wc = L'\0';
    int opt;

    if (tk->kmax && (1 + i + tk->kmax < l)) {
      wc = word[1 + i + tk->kmax];
      word[1 + i + tk->kmax] = L'\0';
      opt = 0;
    }
    else {
      opt = PATRICIA_ADD_OPTION_END;
    }

    patricia_add_context_init(&ctx, tk->global, word + i, opt, NULL);
    patricia_add(&ctx, trk_count);

    if (tk->positional) {
      patricia_add_context_init(&ctx, tk->positional[i], word + i, opt, NULL);
      patricia_add(&ctx, trk_count);
    }

    if (!opt) {
      word[1 + i + tk->kmax] = wc;
    }
  }
}
